#ifndef SBCLIB_BITMAPFONTS
// Hammer Era bitmapped fonts (Stephen Brooks 2001,02)
// All globals in here are prefixed by "BF_"

/* Preprocessor switches usable with this library:
#define SBCLIB_BITMAPFONTS_ASYNC_RENDERING */

#ifdef SBCLIB_GRAPHICS_DX
#include <graphics_dx.c>
#else
#ifdef SBCLIB_GRAPHICS_VIRTUAL
#include <graphics_dx.c>
#else // SBCLIB_GRAPHICS_OPENGL
#include <graphics.c>
#include <rnd/enclosingpoweroftwo.c>
#endif
#endif
#include <lset.c>
#include <palops.c>
#include <report.c>
#include <rnd/macros.c>
#include <rnd/Max.c>
#include <rnd/Maxi.c>
#include <rnd/Mini.c>
#include <rnd/streq.c>
#include <rnd/streqi.c>
#include <rnd/noneg.c>
#include <rnd/eithertime.c>
#include <rnd/rnd.c>
#include <rnd/striseg.c>
#include <rnd/c2i.c>
#include <rnd/filedata.c>

struct BF_font_tag;

typedef struct
{
	struct BF_font_tag *from;
	float xscale,yscale,skew,boldness;
	int render_c,render_i;
} BF_variation;

BF_variation BF_variation_none={NULL, 1,1,0,0, 0,0},BF_curvar;

#ifdef __linux__
int BF_variation_equal(const BF_variation v,const BF_variation w)
#else
int operator ==(const BF_variation v,const BF_variation w)
#endif
{
	return !memcmp(&v,&w,sizeof(BF_variation));
}

typedef struct BF_font_tag
{
	#ifdef SBCLIB_GFX_D3D8
	IDirect3DTexture8 *letter;
	unsigned lox[512],lxs[512],txs;
	#elif defined(SBCLIB_GRAPHICS_OPENGL)
	GLuint letter;
	unsigned lox[512],loy[512],lxs[512],txs,tys;
	#else
	unsigned char *letter[512];
	#endif
	float border; // Number of texels "border" on each letter (drawn outside the nominal letter size)
	unsigned width,height,monospacing; // Calculated by BF_fontlookups(BF_font *)
	char *name; // The "height" field under OpenGL refers to the texture not the height of the font variation
	BF_variation v;
} BF_font;

LSet BF_fontset; BF_font **BF_array=NULL;
BF_font *BF_curfont=NULL;
unsigned BF_width,BF_height,BF_monospacing;
float BF_spacmul=0.1; // 0.11 was equivalent to old 0.16 for font_dat
	//0.16/*0.13*/; // Old when fraction of monospacing, which was 'C' width+1px

#ifndef SBCLIB_GFX_D3D8
#ifndef SBCLIB_GRAPHICS_OPENGL
unsigned char **BF_letter;
#endif
#endif

char BF_mono=0; int BF_leftclipmargin=0;
enum BF_painttype {Solid=1,And=2,Or=3,Alpha=4} BF_paint=Or;
int BF_shaderprog=0;

#ifndef SBCLIB_GRAPHICS_OPENGL
unsigned char BF_and[255][255],BF_or[255][255],BF_inv[255],BF_cra_loaded=0;

void BF_make_cra(void)
{
	int n,i; FILE *in;
	if (BF_cra_loaded) return;
	if (in=fopen("crarray.cgf","rb"))
	{
		unsigned char cra[256][255];
		for (n=0;n<256;n++) fread(cra[n],1,255,in);
		for (n=0;n<255;n++) for (i=0;i<255;i++)
		{
			BF_and[n][i]=cra[Mini(n,i)][Maxi(n,i)];
			BF_or[n][i]=cra[Maxi(n,i)+1][Mini(n,i)];
		}
		fclose(in);
	}
	else if (cpp==1)
	{
		float r[256],g[256],b[256];
		unsigned char cra[256][255]; long lastrefresh=GetTickCount();
		fpalcoeffs(r,g,b);
		for (n=0;n<255;n++)
		{
			for (i=0;i<=n;i++)
			{
				float rr=r[n]*r[i],gg=g[n]*g[i],bb=b[n]*b[i],bd=1e6,dd;
				unsigned char c,cc;
				for (c=0;c<40;c++) if ((dd=sq(rr-r[c])+sq(gg-g[c])+sq(bb-b[c]))<bd) {bd=dd; cc=c;}
				c=40+(int)floor(rr*5+0.5)+(int)floor(gg*5+0.5)*6+(int)floor(bb*5+0.5)*36;
			    if ((dd=sq(rr-r[c])+sq(gg-g[c])+sq(bb-b[c]))<bd) {bd=dd; cc=c;}
				cra[i][n]=BF_and[n][i]=BF_and[i][n]=cc;
				rr=1.0-(1.0-r[n])*(1.0-r[i]); gg=1.0-(1.0-g[n])*(1.0-g[i]); bb=1.0-(1.0-b[n])*(1.0-b[i]); bd=1e6;
				for (c=0;c<40;c++) if ((dd=sq(rr-r[c])+sq(gg-g[c])+sq(bb-b[c]))<bd) {bd=dd; cc=c;}
				c=40+(int)floor(rr*5+0.5)+(int)floor(gg*5+0.5)*6+(int)floor(bb*5+0.5)*36;
				if ((dd=sq(rr-r[c])+sq(gg-g[c])+sq(bb-b[c]))<bd) cc=c;
				cra[n+1][i]=BF_or[n][i]=BF_or[i][n]=cc;
			}
			if (shad && GetTickCount()-lastrefresh>200)
			{
				line(0,yres/2,xres-1,yres/2,42);
				line(0,yres/2,xres*sq(n)/sq(255),yres/2,45);
				ccl(0); lastrefresh=GetTickCount();
			}
		}
		in=fopen("crarray.cgf","wb");
		for (n=0;n<256;n++) fwrite(cra[n],1,255,in);
		fclose(in);
	}
	if (cpp==1) // Colour inverses
	{
		float r[256],g[256],b[256],rr,gg,bb,bd,dd;
		fpalcoeffs(r,g,b);
		unsigned char c,cc;
		for (n=0;n<255;n++)
		{
			rr=1.0-r[n]; gg=1.0-g[n]; bb=1.0-b[n]; bd=1e6; cc=0;
			for (c=0;c<40;c++) if ((dd=sq(rr-r[c])+sq(gg-g[c])+sq(bb-b[c]))<bd) {bd=dd; cc=c;}
			c=40+(int)floor(rr*5+0.5)+((int)floor(gg*5+0.5))*6+((int)floor(bb*5+0.5))*36;
			if ((dd=sq(rr-r[c])+sq(gg-g[c])+sq(bb-b[c]))<bd) cc=c;
			BF_inv[n]=cc;
		}
	}
	BF_cra_loaded=1;
}

void wrapbitcr(int x,int y,const unsigned char *bitmap,const unsigned char *cr)
{
	unsigned xs=bitmap[0],ys=bitmap[1],px=(x<0)?(x+=xres):0,py=(y<0)?(y+=yres):0;
	const unsigned char *c=bitmap+2;
	unsigned char *p=shad+y*shadpitch+x,*l=shad+(yres-1)*shadpitch+xres-xs;
	for (py=0;py<ys;py++)
	{
		if (p>=l) p-=shadpitch*(yres-1);
		for (px=0;px<xs;px++,c++,p++) if (*c!=255) *p=cr[*c];
		p+=shadpitch-xs;
	}
}

void wrapbitc(const int x,const int y,const unsigned char *bitmap,const unsigned col)
{
	if (cpp==1) wrapbitcr(x,y,bitmap,BF_and[col]);
	else
	{
		int x1=x,y1=y;
		unsigned xs=bitmap[0],ys=bitmap[1],px=(x1<0)?x1+=xres:0,py=(y1<0)?y1+=yres:0;
		const unsigned char *c=bitmap+2;
		unsigned char *p=shad+y1*shadpitch+x1*cpp,*l=shad+(yres-1)*shadpitch+(xres-xs)*cpp;
		int colr=col>>16,colg=(col>>8)&0xFF,colb=col&0xFF;
		for (py=0;py<ys;py++)
		{
			if (p>=l) p-=shadpitch*(yres-1);
			for (px=0;px<xs;px++)
			{
				if (c[0]!=0x56 || c[1]!=0x34 || c[2]!=0x12)
				{
					p[0]=(colb*c[0])>>8;
					p[1]=(colg*c[1])>>8;
					p[2]=(colr*c[2])>>8;
				}
				p+=cpp; c+=3;
			}
			p+=shadpitch-xs*cpp;
		}
	}
}

void orbit(int x,int y,const unsigned char *bitmap)
{
	if (cpp==1)
	{
		unsigned xs=bitmap[0],ys=bitmap[1],px=(x<xmin)?x+=xres-xmin:0,py=(y<ymin)?y+=yres-ymin:0;
		const unsigned char *c=bitmap+2;
		unsigned char d,*p=shad+y*shadpitch+x,*l=shad+(yres-1)*shadpitch+xres-xs;
		for (py=0;py<ys;py++)
		{
			if (p>=l) p-=(yres-1)*shadpitch;
			for (px=0,d=*c;px<xs;px++,c++,p++,d=*c) if (d && d!=255) *p=BF_or[*p][d];
			p+=shadpitch-xs;
		}
	}
	else
	{
		int x1=x,y1=y;
		unsigned xs=bitmap[0],ys=bitmap[1],px=(x1<xmin)?x1+=xres-xmin:0,py=(y1<ymin)?y1+=yres-ymin:0;
		const unsigned char *c=bitmap+2;
		unsigned char *p=shad+y1*shadpitch+x1*cpp,*l=shad+(yres-1)*shadpitch+(xres-xs)*cpp;
		for (py=0;py<ys;py++)
		{
			if (p>=l) p-=shadpitch*(yres-1);
			for (px=0;px<xs;px++)
			{
				if (c[0]!=0x56 || c[1]!=0x34 || c[2]!=0x12)
				{
					p[0]=0xFF^(((0xFF^p[0])*(0xFF^c[0]))>>8);
					p[1]=0xFF^(((0xFF^p[1])*(0xFF^c[1]))>>8);
					p[2]=0xFF^(((0xFF^p[2])*(0xFF^c[2]))>>8);
				}
				p+=cpp; c+=3;
			}
			p+=shadpitch-xs*cpp;
		}
	}
}

void orbitcr(int x,int y,const unsigned char *bitmap,const unsigned char *cr)
{
	unsigned xs=bitmap[0],ys=bitmap[1],px=(x<xmin)?(x+=xres-xmin):0,py=(y<ymin)?(y+=yres-ymin):0;
	const unsigned char *c=bitmap+2;
	unsigned char d,*p=shad+y*shadpitch+x,*l=shad+yres*shadpitch-xs;
	for (py=0;py<ys;py++)
	{
		for (px=0,d=*c;px<xs;px++,c++,p++,d=*c) if (d && d!=255) *p=BF_or[*p][cr[d]];
		p+=shadpitch-xs;
		if (p>=l) p-=shadpitch*(yres-1);
	}
}

void orbitc(const int x,const int y,const unsigned char *bitmap,const unsigned col)
{
	if (cpp==1) orbitcr(x,y,bitmap,BF_and[col]);
	else
	{
		int x1=x,y1=y;
		unsigned xs=bitmap[0],ys=bitmap[1],px=(x1<xmin)?x1+=xres-xmin:0,py=(y1<ymin)?y1+=yres-ymin:0;
		const unsigned char *c=bitmap+2;
		unsigned char *p=shad+y1*shadpitch+x1*cpp,*l=shad+(yres-1)*shadpitch+(xres-xs)*cpp;
		int colr=col>>16,colg=(col>>8)&0xFF,colb=col&0xFF;
		for (py=0;py<ys;py++)
		{
			if (p>=l) p-=shadpitch*(yres-1);
			for (px=0;px<xs;px++)
			{
				if (c[0]!=0x56 || c[1]!=0x34 || c[2]!=0x12)
				{
					p[0]=0xFF^(((0xFF^((colb*c[0])>>8))*(0xFF^p[0]))>>8);
					p[1]=0xFF^(((0xFF^((colg*c[1])>>8))*(0xFF^p[1]))>>8);
					p[2]=0xFF^(((0xFF^((colr*c[2])>>8))*(0xFF^p[2]))>>8);
				}
				p+=cpp; c+=3;
			}
			p+=shadpitch-xs*cpp;
		}
	}
}

void andbitcr(int x,int y,const unsigned char *bitmap,const unsigned char *cr)
{
	unsigned xs=bitmap[0],ys=bitmap[1],px=(x<xmin)?(x+=xres-xmin):0,py=(y<ymin)?(y+=yres-ymin):0;
	const unsigned char *c=bitmap+2;
	unsigned char d,*p=shad+y*shadpitch+x,*l=shad+yres*shadpitch-xs;
	for (py=0;py<ys;py++)
	{
		for (px=0,d=*c;px<xs;px++,c++,p++,d=*c) if (d && d!=255) *p=BF_and[*p][cr[d]];
		p+=shadpitch-xs;
		if (p>=l) p-=shadpitch*(yres-1);
	}
}

void andbitc(const int x,const int y,const unsigned char *bitmap,const unsigned col)
{
	if (cpp==1)
	{
		unsigned char cr[256];
		for (int n=0;n<256;n++) cr[n]=BF_or[col][BF_inv[n]];
		andbitcr(x,y,bitmap,cr);
	}
	else
	{
		int x1=x,y1=y;
		unsigned xs=bitmap[0],ys=bitmap[1],px=(x1<xmin)?x1+=xres-xmin:0,py=(y1<ymin)?y1+=yres-ymin:0;
		const unsigned char *c=bitmap+2;
		unsigned char *p=shad+y1*shadpitch+x1*cpp,*l=shad+(yres-1)*shadpitch+(xres-xs)*cpp;
		int colr=col>>16,colg=(col>>8)&0xFF,colb=col&0xFF;
		for (py=0;py<ys;py++)
		{
			if (p>=l) p-=shadpitch*(yres-1);
			for (px=0;px<xs;px++)
			{
				if (c[0]!=0x56 || c[1]!=0x34 || c[2]!=0x12)
				{
					p[0]=(p[0]*(0xFF^(((0xFF^colb)*c[0])>>8)))>>8;
					p[1]=(p[1]*(0xFF^(((0xFF^colg)*c[1])>>8)))>>8;
					p[2]=(p[2]*(0xFF^(((0xFF^colr)*c[2])>>8)))>>8;
				}
				p+=cpp; c+=3;
			}
			p+=shadpitch-xs*cpp;
		}
	}
}

void alphabitcr(int x,int y,const unsigned char *bitmap,const unsigned char *cr)
{
	unsigned xs=bitmap[0],ys=bitmap[1],px=(x<0)?(x+=xres):0,py=(y<0)?(y+=yres):0;
	const unsigned char *c=bitmap+2;
	unsigned char d,*p=shad+y*shadpitch+x,*l=shad+yres*shadpitch-xs;
	for (py=0;py<ys;py++)
	{
		for (px=0,d=*c;px<xs;px++,c++,p++,d=*c) if (d && d!=255) *p=BF_or[BF_and[*p][BF_inv[d]]][cr[d]];
		p+=shadpitch-xs;
		if (p>=l) p-=shadpitch*(yres-1);
	}
}

void alphabitc(const int x,const int y,const unsigned char *bitmap,const unsigned col)
{
	if (cpp==1) alphabitcr(x,y,bitmap,BF_and[col]);
	else
	{
		int x1=x,y1=y;
		unsigned xs=bitmap[0],ys=bitmap[1],px=(x1<xmin)?x1+=xres-xmin:0,py=(y1<ymin)?y1+=yres-ymin:0;
		const unsigned char *c=bitmap+2;
		unsigned char *p=shad+y1*shadpitch+x1*cpp,*l=shad+(yres-1)*shadpitch+(xres-xs)*cpp;
		int colr=col>>16,colg=(col>>8)&0xFF,colb=col&0xFF;
		for (py=0;py<ys;py++)
		{
			if (p>=l) p-=shadpitch*(yres-1);
			for (px=0;px<xs;px++)
			{
				if (c[0]!=0x56 || c[1]!=0x34 || c[2]!=0x12)
				{
					p[0]=0xFF^(((0xFF^((p[0]*(0xFF^c[0]))>>8))*(0xFF^((colb*c[0])>>8)))>>8);
					p[1]=0xFF^(((0xFF^((p[1]*(0xFF^c[1]))>>8))*(0xFF^((colg*c[1])>>8)))>>8);
					p[2]=0xFF^(((0xFF^((p[2]*(0xFF^c[2]))>>8))*(0xFF^((colr*c[2])>>8)))>>8);
				}
				p+=cpp; c+=3;
			}
			p+=shadpitch-xs*cpp;
		}
	}
}

void (*andpixnec)(const unsigned,const unsigned,const unsigned)=NULL,
	(*orpixnec)(const unsigned,const unsigned,const unsigned)=NULL,
	andpixnec1(const unsigned,const unsigned,const unsigned),
	andpixnec3(const unsigned,const unsigned,const unsigned),
	andpixnec4(const unsigned,const unsigned,const unsigned),
	orpixnec1(const unsigned,const unsigned,const unsigned),
	orpixnec3(const unsigned,const unsigned,const unsigned),
	orpixnec4(const unsigned,const unsigned,const unsigned);

void andpix(const unsigned x,const unsigned y,const unsigned c)
{
	if (x>=xmin && x<xres) if (y>=ymin && y<yres) andpixnec(x,y,c);
}

void orpix(const unsigned x,const unsigned y,const unsigned c)
{
	if (c) if (x>=xmin && x<xres) if (y>=ymin && y<yres) orpixnec(x,y,c);
}

void (*paintbitc)(const int,const int,const unsigned char *,const unsigned),
	(*paintpix)(const unsigned,const unsigned,const unsigned);


void BF_setfuncptrs(void)
{
	void bfsetpaint(enum BF_painttype);
	if (cpp==1) {andpixnec=andpixnec1; orpixnec=orpixnec1;}
	else if (cpp==3) {andpixnec=andpixnec3; orpixnec=orpixnec3;}
	else if (cpp==4) {andpixnec=andpixnec4; orpixnec=orpixnec4;}
	bfsetpaint(BF_paint);
}

void andpixnec1(const unsigned x,const unsigned y,const unsigned c)
{
	unsigned char *b=shad+x+y*shadpitch; *b=BF_and[*b][c];
}

void andpixnec3(const unsigned x,const unsigned y,const unsigned c)
{
	unsigned char *b=shad+x*3+y*shadpitch;
	int c0=c&0xFF,c1=(c>>8)&0xFF,c2=c>>16;
	b[0]=((int)b[0]*c0)>>8;
	b[1]=((int)b[1]*c1)>>8;
	b[2]=((int)b[2]*c2)>>8;
}

void andpixnec4(const unsigned x,const unsigned y,const unsigned c)
{
	unsigned char *b=shad+(x<<2)+y*shadpitch;
	int c0=c&0xFF,c1=(c>>8)&0xFF,c2=c>>16;
	b[0]=((int)b[0]*c0)>>8;
	b[1]=((int)b[1]*c1)>>8;
	b[2]=((int)b[2]*c2)>>8;
}

void orpixnec1(const unsigned x,const unsigned y,const unsigned c)
{
	unsigned char *b=shad+x+y*shadpitch;
	*b=BF_or[*b][c];
}

void orpixnec3(const unsigned x,const unsigned y,const unsigned c)
{
	unsigned char *b=shad+x*3+y*shadpitch;
	int c0=0xFF^(c&0xFF),c1=0xFF^((c>>8)&0xFF),c2=0xFF^(c>>16);
	b[0]=0xFF^(((0xFF^b[0])*c0)>>8);
	b[1]=0xFF^(((0xFF^b[1])*c1)>>8);
	b[2]=0xFF^(((0xFF^b[2])*c2)>>8);
}

void orpixnec4(const unsigned x,const unsigned y,const unsigned c)
{
	unsigned char *b=shad+(x<<2)+y*shadpitch;
	int c0=0xFF^(c&0xFF),c1=0xFF^((c>>8)&0xFF),c2=0xFF^(c>>16);
	b[0]=0xFF^(((0xFF^b[0])*c0)>>8);
	b[1]=0xFF^(((0xFF^b[1])*c1)>>8);
	b[2]=0xFF^(((0xFF^b[2])*c2)>>8);
}

void orline(float x1,float y1,float x2,float y2,const unsigned c)
{
	float dx=x2-x1,dy=y2-y1,k,xr=xres-1,yr=yres-1;
	if (x1<xmin)
	{
		if (x2<xmin) return;
		y1-=(x1-xmin)*dy/dx; x1=xmin;
	}
	else if (x2<xmin)
	{
		y2-=(x2-xmin)*dy/dx; x2=xmin;
	}
	if (x1>xr)
	{
		if (x2>xr) return;
		y1-=(x1-xr)*dy/dx; x1=xr;
	}
	else if (x2>xr)
	{
		y2-=(x2-xr)*dy/dx; x2=xr;
	}
	if (y1<ymin)
	{
		if (y2<ymin) return;
		x1-=(y1-ymin)*dx/dy; y1=ymin;
		if (x1<xmin) x1=xmin; else if (x1>xr) x1=xr; // Because of rounding errors in gradient, must check we are not undoing range work on x
	}
	else if (y2<ymin)
	{
		x2-=(y2-ymin)*dx/dy; y2=ymin;
		if (x2<xmin) x2=xmin; else if (x2>xr) x2=xr;
	}
	if (y1>yr)
	{
		if (y2>yr) return;
		x1-=(y1-yr)*dx/dy; y1=yr;
		if (x1<xmin) x1=xmin; else if (x1>xr) x1=xr;
	}
	else if (y2>yr)
	{
		x2-=(y2-yr)*dx/dy; y2=yr;
		if (x2<xmin) x2=xmin; else if (x2>xr) x2=xr;
	}
	dx=x2-x1; dy=y2-y1;
	if (dx*dx+dy*dy<(float)0.1) {orpixnec(x1,y1,c); return;}
	if (dx+dy<(float)0)
	{
		dx=-dx; dy=-dy;
		k=x1; x1=x2; x2=k;
		k=y1; y1=y2; y2=k;
	}
	if (dx>dy)
	{
		k=dy/dx;
		for (;x1<=x2;x1++) {orpixnec(x1,y1,c); y1+=k;}
	}
	else
	{
		k=dx/dy;
		for (;y1<=y2;y1++) {orpixnec(x1,y1,c); x1+=k;}
	}
}

void orbox(const int x,const int y,const int xs,const int ys,const unsigned c)
{
	int n,m;
	if (y>=0 && y<yres) for (n=max(x,0),m=min(x+xs,xres-1);n<=m;n++) orpixnec(n,y,c);
	if (y+ys>=0 && y+ys<yres) for (n=max(x,0),m=min(x+xs,xres-1);n<=m;n++) orpixnec(n,y+ys,c);
	if (x>=0 && x<xres) for (n=max(y+1,0),m=min(y+ys,yres);n<m;n++) orpixnec(x,n,c);
	if (x+xs>=0 && x+xs<xres) for (n=max(y+1,0),m=min(y+ys,yres);n<m;n++) orpixnec(x+xs,n,c);
}
#endif

void bfsetpaint(enum BF_painttype x)
{
	#ifndef SBCLIB_GRAPHICS_OPENGL
	if (x==Solid) {paintbitc=wrapbitc; paintpix=putpix;}
	else if (x==Or) {paintbitc=orbitc; paintpix=orpix;}
	else if (x==And) {paintbitc=andbitc; paintpix=andpix;}
	else if (x==Alpha) {paintbitc=alphabitc; paintpix=putpix;}
	#endif
	BF_paint=x;
}

void BF_initialise(void)
{
	#ifndef SBCLIB_GRAPHICS_OPENGL
	BF_make_cra();
	if (!andpixnec) BF_setfuncptrs();
	#endif
	if (!BF_array) {BF_fontset=LSet(BF_font *); LSet_hookup(&BF_fontset,(void **)&BF_array);}
}

int BF_memusage(void)
{
	int n,c,ret=sizeof(BF_fontset)+BF_fontset.buffer*BF_fontset.elementsize;
	for (n=BF_fontset.m-1;n>=0;n--)
	{
		ret+=sizeof(BF_font)+strlen(BF_array[n]->name);
		#ifdef SBCLIB_GFX_D3D8
		ret+=BF_array[n]->txs*BF_array[n]->height*4;
		#elif defined(SBCLIB_GRAPHICS_OPENGL)
		ret+=BF_array[n]->txs*BF_array[n]->height*3*4/3; // Only an estimate for the mipmap
		#else
		for (c=0;c<512;c++) ret+=2+BF_array[n]->letter[c][0]*BF_array[n]->letter[c][1]*cpp;
		#endif
	}
	return ret;
}

int BF_isfont(const char *handle)
{
	char *base; unsigned n;
	for (n=0;handle[n] && handle[n]!=':';n++);
	base=(char *)malloc(n+1); memcpy(base,handle,n); base[n]=0;
	for (n=0;n<BF_fontset.m;n++) if (!strcmpi(base,BF_array[n]->name)) {free(base); return 1;}
	free(base); return 0;
}

void BF_fontlookups(BF_font *f)
{
	int n;
	#ifdef SBCLIB_GFX_D3D8
	f->width=0; for (n=0;n<512;n++) f->width=max(f->width,f->lxs[n]);
	D3DSURFACE_DESC d;
	f->letter->GetLevelDesc(0,&d);
	f->txs=d.Width; f->height=d.Height;
	f->monospacing=f->lxs['C']+1;
	#elif defined(SBCLIB_GRAPHICS_OPENGL)
	f->width=0; for (n=0;n<512;n++) f->width=max(f->width,f->lxs[n]);
	glBindTexture(GL_TEXTURE_2D,f->letter);
	glGetTexLevelParameteriv(GL_TEXTURE_2D,0,GL_TEXTURE_WIDTH,(GLint *)&f->txs);
	glGetTexLevelParameteriv(GL_TEXTURE_2D,0,GL_TEXTURE_HEIGHT,(GLint *)&f->tys);
	f->monospacing=f->lxs['C']+1;
	#else
	f->width=0; for (n=0;n<512;n++) f->width=max(f->width,f->letter[n][0]);
	f->height=0; for (n=0;n<512;n++) f->height=max(f->height,f->letter[n][1]);
	f->monospacing=f->letter['C'][0]+1;
	#endif
}

void BF_sethandle(BF_font *f,const char *h)
{
	streq(&f->name,h);
}

BF_font *BF_font_copyof(const BF_font *f)
{
	BF_font *ret=(BF_font *)malloc(sizeof(BF_font));
	#ifdef SBCLIB_GFX_D3D8
	d3dev->CreateTexture(f->txs,f->height,1,0,D3DFMT_A8R8G8B8,D3DPOOL_MANAGED,&ret->letter);
	d3dev->CopyRects(f->letter,NULL,1,ret->letter,NULL);
	memcpy(ret->lox,f->lox,512*sizeof(unsigned));
	memcpy(ret->lxs,f->lxs,512*sizeof(unsigned));
	ret->txs=f->txs;
	#elif defined(SBCLIB_GRAPHICS_OPENGL)
	ret->letter=f->letter;
	memcpy(ret->lox,f->lox,512*sizeof(unsigned));
	memcpy(ret->loy,f->loy,512*sizeof(unsigned));
	memcpy(ret->lxs,f->lxs,512*sizeof(unsigned));
	ret->txs=f->txs; ret->tys=f->tys;
	#else
	int n,i;
	for (n=0;n<512;n++)
	{
		i=2+(cpp==1?1:3)*f->letter[n][0]*f->letter[n][1];
		ret->letter[n]=(unsigned char *)malloc(i);
		memcpy(ret->letter[n],f->letter[n],i);
	}
	#endif
	ret->border=f->border;
	ret->width=f->width; ret->height=f->height; ret->monospacing=f->monospacing;
	streqi(&ret->name,f->name);
	ret->v=f->v;
	LSet_add(&BF_fontset,&ret);
	return ret;
}

#include <rnd/Min.c>

float BFF_varfuzzymatch(BF_variation v,BF_variation w)
{ // Returns <=0 for no match, positive indicates extent of match
	if (v.from!=w.from) return -1000;
	float xscalediff=Min(fabs(log(v.xscale/w.xscale))/log(1.05),fabs((v.xscale*v.from->width-w.xscale*w.from->width)/0.5)),
		yscalediff=Min(fabs(log(v.yscale/w.yscale))/log(1.05),fabs((v.yscale*v.from->height-w.yscale*w.from->height)/0.5)),
		scalediff=0.5*(xscalediff+yscalediff),
		bolddiff=fabs((v.boldness-w.boldness)/0.02),
		skewdiff=fabs((v.skew-w.skew)/0.05),
		tdiff=scalediff+bolddiff+skewdiff;
	return 1.0-tdiff;
}

#ifndef SBCLIB_GRAPHICS_OPENGL
float *BFF_intensitymap(unsigned char *a,const int X,const int Y)
{
	float *ret=(float *)malloc(X*Y*sizeof(float)),*rend=ret+X*Y,*b;
	if (cpp==1) for (b=ret;b<rend;a++,b++)
	{
		if (*a==255) *b=0;
		else
		{
			PALETTEENTRY p=pal[*a];
			*b=(float)(3.0*p.peRed+5.0*p.peGreen+2.0*p.peBlue)/2550;
		}
	}
	else for (b=ret;b<rend;a+=3,b++)
	{
		if (((a[2]<<16)|(a[1]<<8)|a[0])==0x123456) *b=0;
		else *b=(float)(3.0*a[2]+5.0*a[1]+2.0*a[0])/2550;
	}
	return ret;
}
#endif

#ifndef SBCLIB_GFX_D3D8 // Unsupported (so far)
#ifndef SBCLIB_GRAPHICS_OPENGL
void BFF_copychar(BF_font *from,const int c,BF_font *to)
{
	if (from==to) return;
	free(to->letter[c]);
	int sz=2+(cpp>=3?3:1)*from->letter[c][0]*from->letter[c][1];
	to->letter[c]=(unsigned char *)malloc(sz);
	memcpy(to->letter[c],from->letter[c],sz);
}

void BFF_rescalechar(BF_font *f,const int c,const float zx,const float zy)
{
	if (zx>1 && zy<1) {BFF_rescalechar(f,c,zx,1); BFF_rescalechar(f,c,1,zy); return;}
	if (zx<1 && zy>1) {BFF_rescalechar(f,c,1,zy); BFF_rescalechar(f,c,zx,1); return;}
	int i,x,y; unsigned char *cc=f->letter[c];
	int xs=cc[0],ys=cc[1],nxs=floor(zx*xs)+1,nys=floor(zy*ys)+1;
	float *glevel=BFF_intensitymap(cc+2,xs,ys),
		*rslevel=(float *)malloc(nxs*nys*sizeof(float));
	free(cc); cc=f->letter[c]=(unsigned char *)malloc(2+(cpp>=3?3:1)*nxs*nys);
	cc[0]=nxs; cc[1]=nys;
	if (zx<=1 && zy<=1) // Pixel scaling (with intensity normalisation)
	{
		for (i=nxs*nys-1;i>=0;i--) rslevel[i]=0;
		for (x=0;x<xs;x++) for (y=0;y<ys;y++)
		{
			int ix=floor(zx*x),iy=floor(zy*y),jx=floor(zx*(x+1)),jy=floor(zy*(y+1));
			float g=glevel[x+y*xs];
			if (ix==jx)
			{
				if (iy==jy) rslevel[ix+iy*nxs]+=g;
				else //if (iy+1==jy)
				{
					float fy=jy/zy-y;
					rslevel[ix+iy*nxs]+=g*fy;
					rslevel[ix+jy*nxs]+=g*(1.0-fy);
				}
			}
			else //if (ix+1==jx)
			{
				if (iy==jy)
				{
					float fx=jx/zx-x;
					rslevel[ix+iy*nxs]+=g*fx;
					rslevel[jx+iy*nxs]+=g*(1.0-fx);
				}
				else
				{
					float fx=jx/zx-x,fy=jy/zy-y;
					rslevel[ix+iy*nxs]+=g*fx*fy;
					rslevel[jx+iy*nxs]+=g*(1.0-fx)*fy;
					rslevel[ix+jy*nxs]+=g*fx*(1.0-fy);
					rslevel[jx+jy*nxs]+=g*(1.0-fx)*(1.0-fy);
				}
			}
		}
		float imax=0.001;
		for (i=nxs*nys-1;i>=0;i--) imax=Max(imax,rslevel[i]);
		for (i=nxs*nys-1;i>=0;i--) rslevel[i]=pow(rslevel[i]/imax,(c>='0' && c<='9' || c>=256 && c<384)?1.0/1.8:1.0); // Gamma hack
		if (cpp==1)
		{ // noneg in here is needed because of rounding errors producing very small negative values in the above (try zx=zy=0.7)
			for (x=0;x<nxs;x++) for (y=0;y<nys;y++)
				cc[2+x+y*nxs]=(int)floor(16.0+noneg(rslevel[x+y*nxs])*23.9); // For some reason this GPFs without the (int) cast!
		}
		else
		{
			unsigned char *p;
			for (x=0;x<nxs;x++) for (y=0;y<nys;y++)
			{
				p=cc+2+(x+y*nxs)*3;
				p[0]=p[1]=p[2]=rslevel[x+y*nxs]*255.9;
			}
		}
	}
	else if (zx>=1 && zy>=1) // Bilinear with cutoff
	{
		int xp=(cpp>=3?3:1),yp=xp*nxs;
		for (x=0;x<nxs;x++) for (y=0;y<nys;y++)
		{
			float ox=(x+0.5)/zx-0.5,oy=(y+0.5)/zy-0.5;
			int ix=floor(ox),iy=floor(oy),jx=ix+1,jy=iy+1;
			float a=0,b=0,c=0,d=0,fx=ox-ix,fy=oy-iy;
			if (ix>=0 && ix<xs)
			{
				if (iy>=0 && iy<ys) a=glevel[ix+iy*xs];
				if (jy>=0 && jy<ys) c=glevel[ix+jy*xs];
			}
			if (jx>=0 && jx<xs)
			{
				if (iy>=0 && iy<ys) b=glevel[jx+iy*xs];
				if (jy>=0 && jy<ys) d=glevel[jx+jy*xs];
			}
			cc[2+x*xp+y*yp]=(a+(b-a)*fx+(c-a+(a-b-c+d)*fx)*fy>0.5)*15;
		}
		int aa=3;
		for (x=0;x<nxs;x++) for (y=0;y<nys;y++) // Multisample antialiasing of bilinear contour
		{
			char edge=0; i=2+x*xp+y*yp;
			if (x-1>=0) if (cc[i-xp]!=cc[i]) edge=1;
			if (x+1<nxs) if (cc[i+xp]!=cc[i]) edge=1;
			if (y-1>=0) if (cc[i-yp]!=cc[i]) edge=1;
			if (y+1<nys) if (cc[i+yp]!=cc[i]) edge=1;
			if (edge)
			{
				cc[i]=0;
				for (float xx=0.5/aa;xx<0.99;xx+=1.0/aa) for (float yy=0.5/aa;yy<0.99;yy+=1.0/aa)
				{
					float ox=(x+xx)/zx-0.5,oy=(y+yy)/zy-0.5;
					int ix=floor(ox),iy=floor(oy),jx=ix+1,jy=iy+1;
					float a=0,b=0,c=0,d=0,fx=ox-ix,fy=oy-iy;
					if (ix>=0)
					{
						if (iy>=0) a=glevel[ix+iy*xs];
						if (jy<ys) c=glevel[ix+jy*xs];
					}
					if (jx<xs)
					{
						if (iy>=0) b=glevel[jx+iy*xs];
						if (jy<ys) d=glevel[jx+jy*xs];
					}
					cc[i]+=(a+(b-a)*fx+(c-a+(a-b-c+d)*fx)*fy)>0.5;
				}
				if (cpp==1) cc[i]=16.0+23.9*((float)cc[i]/aa/aa);
			}
		}
		if (cpp>=3) for (i=nxs*nys-1;i>=0;i--)
				cc[2+i*3]=cc[2+i*3+1]=cc[2+i*3+2]=(unsigned char)(255.9*((float)cc[2+i*3]/aa/aa));
	}
	free(rslevel); free(glevel);
}

void BFF_recanvaschar(BF_font *f,const int c,const unsigned nxs,const unsigned nys)
{ // Leaves char same size, centred on canvas
	int n,i,x,y; unsigned char *cc=f->letter[c];
	int xs=cc[0],ys=cc[1];
	unsigned char *nc=(unsigned char *)malloc(2+(cpp>=3?3:1)*nxs*nys);
	nc[0]=nxs; nc[1]=nys;
	int xp=(cpp>=3?3:1),yp=xp*nxs,xf,yf,yp0=xp*xs;
	for (y=nys-1;y>=0;y--) for (x=nxs-1;x>=0;x--)
	{
		i=2+x*xp+y*yp;
		xf=x-nxs/2+xs/2; yf=y-nys/2+ys/2;
		if (xf>=0 && xf<xs && yf>=0 && yf<ys)
		{
			n=2+xf*xp+yf*yp0;
			if (cpp==1) nc[i]=cc[n];
			else {nc[i]=cc[n]; nc[i+1]=cc[n+1]; nc[i+2]=cc[n+2];}
		}
		else
		{
			if (cpp==1) nc[i]=255;
			else {nc[i]=0x56; nc[i+1]=0x34; nc[i+2]=0x12;}
		}
	}
	free(cc); f->letter[c]=nc;
}

void BFF_boldchar(BF_font *f,const int c,const float boldness)
{
	int i,j,x,y; unsigned char *cc=f->letter[c];
	float boldhw=boldness*f->width/2; int cw=ceil(fabs(boldhw)*2);
	int xs=cc[0],ys=cc[1],nxs=Maxi(1,xs+ceil(boldhw*2));
	float *glevel=BFF_intensitymap(cc+2,xs,ys),
		*rslevel=(float *)malloc(nxs*sizeof(float)),
		*conv=(float *)malloc((cw+1)*sizeof(float));
	free(cc); cc=f->letter[c]=(unsigned char *)malloc(2+(cpp>=3?3:1)*nxs*ys);
	cc[0]=nxs; cc[1]=ys;
	boldhw=fabs(boldhw);
	for (int n=cw;n>=0;n--) conv[n]=1;
	if (boldhw*2!=ceil(boldhw*2))
	{
		if (cw%2) conv[cw]=fmod(boldhw*2,1);
		else conv[0]=fmod(boldhw*2,1);
	}
	float *glevely=glevel;
	for (y=0;y<ys;y++,glevely+=xs)
	{
		if (boldness>=0) // Positive boldness
		{
			for (i=nxs-1;i>=0;i--) rslevel[i]=0;
			for (x=0;x<xs;x++) for (i=0;i<=cw;i++) rslevel[x+i]=Min(1,rslevel[x+i]+glevely[x]*conv[i]);
		}
		else // Oddnegative boldness
		{
			for (i=nxs-1;i>=0;i--) rslevel[i]=1;
			for (x=-cw;x<xs+cw;x++) for (i=0;i<=cw;i++)
			{
				float glev=0;
				if (x>=0 && x<xs) glev=glevely[x];
				j=x+i-cw+((int)(floor(boldhw*2))%2==0);
				if (j>=0 && j<nxs) rslevel[j]=Min(rslevel[j],1.0-(1.0-glev)*conv[i]);
			}
		}
		if (cpp==1) for (x=nxs-1;x>=0;x--) cc[2+x+y*nxs]=(int)floor(16.0+rslevel[x]*23.9);
		else
		{
			unsigned char *p;
			for (x=nxs-1;x>=0;x--)
			{
				p=cc+2+(x+y*nxs)*3;
				p[0]=p[1]=p[2]=rslevel[x]*255.9;
			}
		}
	}
	free(rslevel); free(glevel); free(conv);
}

void BFF_skewchar(BF_font *f,const int c,const float skew)
{
	int i,j,x,y; unsigned char *cc=f->letter[c];
	int xs=cc[0],ys=cc[1];
	float skm=fabs(skew)*(ys-1);
	int nxs=xs+floor(skm+0.75);
	float *glevel=BFF_intensitymap(cc+2,xs,ys),
		*rslevel=(float *)malloc(nxs*sizeof(float));
	free(cc); cc=f->letter[c]=(unsigned char *)malloc(2+(cpp>=3?3:1)*nxs*ys);
	cc[0]=nxs; cc[1]=ys;
	float *glevely=glevel;
	for (y=0;y<ys;y++,glevely+=xs)
	{
		float offset;
		if (skew>=0) offset=skew*(ys-1-y); else offset=-skew*y;
		for (i=nxs-1;i>=0;i--) rslevel[i]=0;
		for (x=0;x<xs;x++)
		{
			i=floor(x+offset); j=Mini(i+1,nxs-1); float p=x+offset-i;
			rslevel[i]+=glevely[x]*(1.0-p);
			rslevel[j]+=glevely[x]*p;
		}
		if (cpp==1) for (x=nxs-1;x>=0;x--) cc[2+x+y*nxs]=(int)floor(16.0+rslevel[x]*23.9);
		else
		{
			unsigned char *p;
			for (x=nxs-1;x>=0;x--)
			{
				p=cc+2+(x+y*nxs)*3;
				p[0]=p[1]=p[2]=rslevel[x+y*nxs]*255.9;
			}
		}
	}
	free(rslevel); free(glevel);
}

void BFF_trimchar(BF_font *f,const int c,const float cutoff)
{
	int x,y; unsigned char *cc=f->letter[c];
	int xs=cc[0],ys=cc[1],nxs=xs,nys=ys,ox=0,oy=0;
	float *glevel=BFF_intensitymap(cc+2,xs,ys);
	char flag=1;
	for (x=xs*ys-1;x>=0;x--) if (glevel[x]<=cutoff)
	{
		if (cpp==1) cc[2+x]=255;
		else {cc[2+x*3]=0x56; cc[2+x*3+1]=0x34; cc[2+x*3+2]=0x12;}
	} else flag=0;
	if (flag) return; // Don't crop spacers
	for (char go=1;go && nxs>1;) // Right
	{
		for (y=ys-1;y>=0;y--) if (glevel[nxs-1+y*xs]>cutoff) {go=0; y=-1;}
		if (go) nxs--;
	}
	for (char go=1;go && nxs>1;) // Left
	{
		for (y=ys-1;y>=0;y--) if (glevel[ox+y*xs]>cutoff) {go=0; y=-1;}
		if (go) {ox++; nxs--;}
	}
	for (char go=1;go && nys>1;) // Bottom
	{
		for (x=xs-1;x>=0;x--) if (glevel[x+(nys-1)*xs]>cutoff) {go=0; x=-1;}
		if (go) nys--;
	}
	free(glevel);
	unsigned char *nc=f->letter[c]=(unsigned char *)malloc(2+(cpp>=3?3:1)*nxs*nys);
	nc[0]=nxs; nc[1]=nys;
	if (cpp==1) for (y=0;y<nys;y++) memcpy(nc+2+y*nxs,cc+2+ox+(y+oy)*xs,nxs);
	else for (y=0;y<nys;y++) memcpy(nc+2+y*nxs*3,cc+2+(ox+(y+oy)*xs)*3,nxs*3);
	free(cc);
}

void BFF_trimfont(BF_font *f=NULL,const float cutoff=0)
{
	if (!f) f=BF_curfont;
	for (int c=0;c<512;c++) BFF_trimchar(f,c,cutoff);
	BF_fontlookups(f);
}

int BFF_render_index=0;

void BFF_render(const float ms)
{
	double endtime=eithertime()+0.001*ms,tt;
	#ifndef SBCLIB_BITMAPFONTS_ASYNC_RENDERING
	endtime=9e99;
	#endif
	BFF_render_index%=BF_fontset.m;
	int origindex=BFF_render_index;
	char alldone=0,changed=0;
	while ((tt=eithertime())<endtime && !alldone)
	{
		BF_font *f=BF_array[BFF_render_index];
		if (f->v.render_c<512)
		{
			int c=(f->v.render_c*81)&511,i=f->v.render_i;
			switch (i)
			{
				case 0: BFF_copychar(f->v.from,c,f);
					if (f->v.xscale!=1 || f->v.yscale!=1) BFF_rescalechar(f,c,f->v.xscale,f->v.yscale);
					break;
				case 1: if (f->v.boldness!=0) BFF_boldchar(f,c,f->v.boldness); break;
				case 2: if (f->v.skew!=0) BFF_skewchar(f,c,f->v.skew); break;
				case 3: BFF_trimchar(f,c,0.25);
			}
			if (i<3) f->v.render_i++;
			else {f->v.render_i=0; f->v.render_c++; if (!(f->v.render_c&31)) BF_fontlookups(f);}
			changed=1;
		}
		BFF_render_index=(BFF_render_index+1)%BF_fontset.m;
		if (BFF_render_index==origindex) {if (!changed) alldone=1; changed=0;}
	}
}
#endif

BF_font *bffontvar(BF_variation var)
{
	int n; BF_font *ret=NULL,*bma=var.from; float bestmatch=0,bestmatchany=-1e20;
	for (n=BF_fontset.m-1;n>=0;n--)
	{
		float match=BFF_varfuzzymatch(var,BF_array[n]->v);
		if (match>bestmatch) {ret=BF_array[n]; bestmatch=match;}
		if (match>bestmatchany && BF_array[n]->v.render_c==512) {bma=BF_array[n]; bestmatchany=match;}
	}
	if (ret) return ret;
	ret=BF_font_copyof(bma->v.from);
	var.render_c=var.render_i=0; ret->v=var;
	if (var.xscale!=1) {char str[1000]; sprintf(str,"%s:width=%g",ret->name,var.xscale); streq(&ret->name,str);}
	if (var.yscale!=1) {char str[1000]; sprintf(str,"%s:height=%g",ret->name,var.yscale); streq(&ret->name,str);}
	if (var.boldness!=0) {char str[1000]; sprintf(str,"%s:boldness=%g",ret->name,var.boldness); streq(&ret->name,str);}
	if (var.skew!=0) {char str[1000]; sprintf(str,"%s:skew=%g",ret->name,var.skew); streq(&ret->name,str);}
	#ifndef SBCLIB_GRAPHICS_OPENGL
	BFF_render(0);
	#else
	BF_fontlookups(ret);
	#endif
	return ret;
}
#endif

void bfsetfontpointer(BF_font *f)
{
	if (BF_curfont==f || !f) return;
	BF_curfont=f; BF_curvar=f->v;
	#ifndef SBCLIB_GFX_D3D8
	#ifndef SBCLIB_GRAPHICS_OPENGL
	BF_letter=f->letter;
	#endif
	#endif
	#ifdef SBCLIB_GRAPHICS_OPENGL
	BF_width=(unsigned)(f->v.from->width*f->v.xscale);
	BF_height=(unsigned)(f->v.from->height*f->v.yscale);
	BF_monospacing=(unsigned)(f->v.from->monospacing*f->v.xscale);
	#else
	BF_width=f->width;
	BF_height=f->height;
	BF_monospacing=f->monospacing;
	#endif
}

void BF_deletefont(BF_font *f)
{
	int n;
	#ifdef SBCLIB_GFX_D3D8
	f->letter->Release();
	#elif defined(SBCLIB_GRAPHICS_OPENGL)
	for (n=BF_fontset.m-1;n>=0;n--) if (BF_array[n]!=f && BF_array[n]->letter==f->letter) n=-10;
	if (n>-10) glDeleteTextures(1,&f->letter);
	#else
	for (n=0;n<512;n++) free(f->letter[n]);
	#endif
	free(f->name); free(f);
	LSet_delete(&BF_fontset,&f);
	if (BF_curfont==f) bfsetfontpointer(BF_array[rnd(BF_fontset.m)]);
}

void BF_deletefontvariants(BF_font *f)
{
	int n;
	for (n=BF_fontset.m-1;n>=0;n--)
		if (BF_array[n]->v.from==f) BF_deletefont(BF_array[n]);
}

BF_font *BF_makeblank(const char *handle)
{
	int n,i; BF_font *f;
	BF_initialise();
	for (n=0;n<BF_fontset.m;n++) if (!strcmpi(BF_array[n]->name,handle)) BF_deletefont(BF_array[n]);
	f=(BF_font *)mallocex(sizeof(BF_font),"BF_font structure");
	f->name=(char *)mallocex(strlen(handle)+1,"Handle to BF_font");
	LSet_add(&BF_fontset,&f);
	strcpy(f->name,handle);
	f->border=0;
	#ifdef SBCLIB_GFX_D3D8
	d3dev->CreateTexture(f->txs=1,1,1,0,D3DFMT_A8R8G8B8,D3DPOOL_MANAGED,&f->letter);
	for (i=0;i<512;i++) {f->lox[n]=0; f->lxs[n]=1;}
	#elif defined(SBCLIB_GRAPHICS_OPENGL)
	glGenTextures(1,&f->letter);
	glBindTexture(GL_TEXTURE_2D,f->letter);
	glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_LOD_BIAS,-0.5);
	unsigned char *td=(unsigned char *)malloc(1); td[0]=1;
	glTexImage2D(GL_TEXTURE_2D,0,GL_RGB,f->txs=1,1,0,GL_RGB,GL_UNSIGNED_BYTE,td);
	free(td);
	for (i=0;i<512;i++) {f->lox[n]=0; f->lxs[n]=1;}
	#else
	for (i=0;i<512;i++)
	{
		if (cpp==1)
		{
			f->letter[i]=(unsigned char *)mallocex(3,"Letter in blank font (8bit)");
			f->letter[i][0]=f->letter[i][1]=1; f->letter[i][2]=1;
		}
		else
		{
			f->letter[i]=(unsigned char *)mallocex(5,"Letter in blank font (24bit)");
			f->letter[i][0]=f->letter[i][1]=1; f->letter[i][2]=f->letter[i][3]=0; f->letter[i][4]=128;
		}
	}
	#endif
	BF_fontlookups(f);
	return f;
}

#ifndef SBCLIB_GFX_D3D8
#ifndef SBCLIB_GRAPHICS_OPENGL
void BF_monospacedigits(BF_font *f)
{
	int n,w=1;
	for (n='0';n<='9';n++) w=max(w,f->letter[n][0]);
	for (n='0';n<='9';n++) BFF_recanvaschar(f,n,w,f->letter[n][1]);
}
#endif

BF_variation bfvarfromhandle(const char *handle) // This parses the request string
{
	int n,i;
	BF_variation var=BF_variation_none; char dlprev=0,dl=1;
	for (n=0;dl;n++) // The n++ skips the pointer over delimiters
	{
		int on=n-1;
		for (;handle[n]!=':' && handle[n];n++);
		dlprev=dl; dl=handle[n];
		if (dlprev==1) // The base font name
		{
			char *h2; streqi(&h2,handle);
			h2[n]=0;
			for (i=0;i<BF_fontset.m && !var.from;i++)
				if (!strcmpi(h2,BF_array[i]->name)) var=BF_array[i]->v;
			free(h2);
			if (!var.from) // If nothing can be found, return the current font's base.  If that doesn't exist, return a random font, or a blank cardboard beak if none are loaded
			{
				if (BF_curfont) var.from=BF_curfont->v.from;
				else if (BF_fontset.m) var.from=BF_array[rnd(BF_fontset.m)]->v.from;
				else var.from=BF_makeblank(handle);
			}
		}
		else if (dlprev==':') // A derived font modification
		{
			char *attrib,*value=NULL;
			for (i=on+1;i<n && handle[i]!='=';i++);
			if (i<n) {value=(char *)malloc(n-i); if (n-i-1>0) memcpy(value,handle+i+1,n-i-1); value[n-i-1]=0;}
			attrib=(char *)malloc(i-on); if (i-on-1>0) memcpy(attrib,handle+on+1,i-on-1); attrib[i-on-1]=0;
			if (!strcmpi(attrib,"normal")) {BF_font *tf=var.from; var=BF_variation_none; var.from=tf;}
			if (!strcmpi(attrib,"scale") || !strcmpi(attrib,"size"))
			{
				double v=atof(value);
				if (strstr(value,"px")) var.xscale=var.yscale=v/var.from->height;
				else
				{
					if (strstr(value,"%")) v/=100;
					var.xscale*=v; var.yscale*=v;
				}
			}
			else if (!strcmpi(attrib,"big")) {var.xscale*=1.5; var.yscale*=1.5;}
			else if (!strcmpi(attrib,"small")) {var.xscale/=1.5; var.yscale/=1.5;}
			else if (!strcmpi(attrib,"width"))
			{
				double v=atof(value);
				if (strstr(value,"px")) var.xscale=v/var.from->height;
				else
				{
					if (strstr(value,"%")) v/=100;
					var.xscale*=v;
				}
			}
			else if (!strcmpi(attrib,"height"))
			{
				double v=atof(value);
				if (strstr(value,"px")) var.yscale=v/var.from->height;
				else
				{
					if (strstr(value,"%")) v/=100;
					var.yscale*=v;
				}
			}
			else if (!strcmpi(attrib,"boldness"))
			{
				double v=atof(value);
				if (strstr(value,"px")) var.boldness+=v/var.from->width;
				else
				{
					if (strstr(value,"%")) v/=100;
					var.boldness+=v;
				}
			}
			else if (!strcmpi(attrib,"demibold")) var.boldness+=0.03;
			else if (!strcmpi(attrib,"bold")) {var.boldness+=0.05; var.xscale*=1.1;}
			else if (!strcmpi(attrib,"black")) {var.boldness+=0.08; var.xscale*=1.2;}
			else if (!strcmpi(attrib,"skew") || !strcmpi(attrib,"angle"))
			{
				double v=atof(value);
				if (strstr(value,"px")) var.skew+=v/var.from->height;
				else if (strstr(value,"\xB0")) var.skew=tan(v degrees);
				else
				{
					if (strstr(value,"%")) v/=100;
					var.skew+=v;
				}
			}
			else if (!strcmpi(attrib,"italic")) var.skew+=0.4;
			free(attrib); if (value) free(value);
		}
	}
	return var;
}

void bfsetfontvar(const BF_variation v)
{
	if (!memcmp(&v,&BF_curvar,sizeof(BF_variation))) return;
	bfsetfontpointer(bffontvar(v));
	BF_curvar=v;
}
#endif

BF_font *bffont(const char *handle)
{
	#ifdef SBCLIB_GFX_D3D8
	for (int n=BF_fontset.m-1;n>=0;n--)
		if (!strcmpi(handle,BF_array[n]->name)) return BF_array[n];
	return NULL;
	#else
	return bffontvar(bfvarfromhandle(handle));
	#endif
}

void bfsetfont(const char *handle)
{
	if (BF_curfont) if (!strcmpi(handle,BF_curfont->name)) return;
	#ifdef SBCLIB_GFX_D3D8
	bfsetfontpointer(bffont(handle));
	#else
	bfsetfontvar(bfvarfromhandle(handle));
	#endif
}

void bfrename(const char *h1,const char *h2)
{
	streq(&bffont(h1)->name,h2);
}

void bfmakecopy(const char *h1,const char *h2)
{
	BF_font *f=BF_font_copyof(bffont(h1));
	BF_sethandle(f,h2);
	#ifndef SBCLIB_GRAPHICS_OPENGL
	f->v.from=f;
	f->v.xscale=f->v.yscale=1; f->v.skew=f->v.boldness=0;
	#endif
}

#ifdef SBCLIB_GRAPHICS_OPENGL
void correctmipmapgamma(const BF_font *f)
{ // OpenGL version of the "gamma hack"
	int n,n0,n1,y,c,lev; unsigned char *map=(unsigned char *)malloc(256);
	for (n=0;n<256;n++) map[n]=(unsigned char)(255.9*pow((double)n/255.9,1.0/2.5));
	glBindTexture(GL_TEXTURE_2D,f->letter);
	for (lev=1;;lev++)
	{ // I tried TexSubImage etc. to speed this up and it didn't
		int xs,ys;
		glGetTexLevelParameteriv(GL_TEXTURE_2D,lev,GL_TEXTURE_WIDTH,&xs);
		glGetTexLevelParameteriv(GL_TEXTURE_2D,lev,GL_TEXTURE_HEIGHT,&ys);
		unsigned char *d=(unsigned char *)malloc(xs*ys*4);
		glGetTexImage(GL_TEXTURE_2D,lev,GL_RGBA,GL_UNSIGNED_BYTE,d);
		for (c=0;c<512;c++) if (c>='0' && c<='9' || c>=256 && c<384)
		{
			float imax=0.001;
			for (y=f->loy[c]>>lev;y<(f->loy[c]+f->height)>>lev;y++)
			{
				n0=((f->lox[c]>>lev)+y*xs)*4;
				n1=n0+(f->lxs[c]>>lev)*4;
				for (n=n0;n<n1;n++) imax=Max(imax,d[n]);
			}
			for (y=f->loy[c]>>lev;y<(f->loy[c]+f->height)>>lev;y++)
			{
				n0=((f->lox[c]>>lev)+y*xs)*4;
				n1=n0+(f->lxs[c]>>lev)*4;
				for (n=n0;n<n1;n++) //d[n]=(d[n]?255:0);
					//d[n]=(d[n]?127.7+127.7*d[n]/imax:0);
					//d[n]=floor(255.9*pow(d[n]/imax,1.0/2.5)); // Needs to be stronger sauce to cope with mipmap blurriness
					d[n]=map[(int)(255.4*d[n]/imax)];
			}
		}
		glTexImage2D(GL_TEXTURE_2D,lev,4,xs,ys,0,GL_RGBA,GL_UNSIGNED_BYTE,d);
		free(d);
		if (xs==1 && ys==1) break;
	}
	free(map);
}

#include <rnd/sjbbuild2dmipmaps.c>
#endif

void loadfontdata(const unsigned char *d,const unsigned dm,const char *handle,const char alphablend=0)
{ // Must be first call (or loadfont), to initialise
	int n,x,y; BF_font *f;
	#ifdef SBCLIB_DEBUG
	char desc[1000]; sprintf(desc,"BF_font '%s'",handle);
	#endif
	f=(BF_font *)mallocex(sizeof(BF_font),desc);
	f->name=(char *)mallocex(strlen(handle)+1,"BF_font handle");
	strcpy(f->name,handle);
	f->border=0; // Default setting, *not* set by BF_fontlookups
	unsigned dp=0;
	#if defined(SBCLIB_GFX_D3D8) || defined(SBCLIB_GRAPHICS_OPENGL)
	int fh=1;
	for (n=0;n<512 && dp<dm;n++)
	{
		x=d[dp++]; y=d[dp++];
		dp+=x*y;
		if (y>fh) fh=y;
	}
	dp=0; f->height=fh; // BF_height set later on
	int ox=0,oy=0,txmax=1,gap=fh/4; // Gap to stop mipmapping causing the letters to leak into each other
	#ifdef SBCLIB_GRAPHICS_OPENGL
	GLint maxtx; glGetIntegerv(GL_MAX_TEXTURE_SIZE,&maxtx);
	#else
	int maxtx=2048;
	#endif
	if (maxtx>2048) maxtx=2048; // For glserver compatibility with laptop L
	for (n=0;n<512;n++)
	{
		f->lox[n]=ox; f->loy[n]=oy;
		if (dp<dm)
		{
			f->lxs[n]=x=d[dp++]; y=d[dp++];
			dp+=x*y;
			ox+=x+gap;
			if (ox>maxtx) {f->lox[n]=0; ox=x+gap; oy+=fh+gap; f->loy[n]=oy;}
			txmax=Maxi(txmax,f->lox[n]+x);
		}
		else f->lxs[n]=0;
	}
	dp=0;
	f->txs=enclosingpoweroftwo(txmax); f->tys=enclosingpoweroftwo(oy+fh);
	if (f->tys>maxtx)
	{
		char str[200];
		sprintf(str,"Font '%s' required texture of size %dx%d but maxtx=%d",handle,f->txs,f->tys,maxtx);
		reportfatala(str);
	}
	#ifdef SBCLIB_GFX_D3D8
	d3dev->CreateTexture(f->txs=ox,fh,1,0,D3DFMT_A8R8G8B8,D3DPOOL_MANAGED,&f->letter);
	D3DLOCKED_RECT tr;
	f->letter->LockRect(0,&tr,NULL,0);
	unsigned char *ta=tr.pBits;
	#else
	unsigned char *ta=(unsigned char *)calloc(f->txs*f->tys,4); // All components (incl. alpha) zero outside text area
	#endif
	#endif
	float r[256],g[256],b[256]; paldata666(r,g,b);
	for (n=0;n<512;n++)
	{
		if (dp<dm)
		{
			x=d[dp++]; y=d[dp++];
			#if defined(SBCLIB_GFX_D3D8) || defined(SBCLIB_GRAPHICS_OPENGL)
			const unsigned char *dl; unsigned char *tl,c;
			for (oy=0;oy<y;oy++)
			{
				dl=d+dp+oy*x;
				#ifdef SBCLIB_GRAPHICS_D3D8
				tl=ta+oy*tr.Pitch+f->lox[n]*4;
				#else // OpenGL
				tl=ta+((f->loy[n]+oy)*f->txs+f->lox[n])*4;
				#endif
				if (alphablend) for (ox=0;ox<x;ox++)
				{ // alphablend uses alpha channel for solidity (monochrome)
					c=dl[ox];
					tl[ox*4]=0xFF;
					tl[ox*4+1]=0xFF;
					tl[ox*4+2]=0xFF;
					tl[ox*4+3]=(unsigned char)(c==0xFF?0:(r[c]+g[c]+b[c])*85.1);
				}
				else for (ox=0;ox<x;ox++)
				{ // Full-color non-transparent, except 0xFF is fully transparent (black)
					c=dl[ox];
					if (c==0xFF)
					{
						tl[ox*4]=0;
						tl[ox*4+1]=0;
						tl[ox*4+2]=0;
						tl[ox*4+3]=0;
					}
					else
					{
						tl[ox*4]=(unsigned char)(r[c]*255.9);
						tl[ox*4+1]=(unsigned char)(g[c]*255.9);
						tl[ox*4+2]=(unsigned char)(b[c]*255.9);
						tl[ox*4+3]=0xFF;
					}
				}
			}
			dp+=x*y;
			#else
			if (cpp==1)
			{
				#ifdef SBCLIB_DEBUG
				char desc[100]; sprintf(desc,"%s ASCII %d (8bit)",handle,n);
				#endif
				f->letter[n]=(unsigned char *)mallocex(2+x*y,desc);
				f->letter[n][0]=x; f->letter[n][1]=y;
				memcpy(f->letter[n]+2,d+dp,x*y); dp+=x*y;
			}
			else
			{
				int px,py; unsigned char *dd,c;
				#ifdef SBCLIB_DEBUG
				char desc[100]; sprintf(desc,"%s ASCII %d (24bit)",handle,n);
				#endif
				f->letter[n]=(unsigned char *)mallocex(2+x*y*3,desc);
				f->letter[n][0]=x; f->letter[n][1]=y;
				dd=f->letter[n]+2;
				for (py=0;py<y;py++) for (px=0;px<x;px++)
				{
					c=d[dp++];
					if (c==0xFF) {*(dd++)=0x56; *(dd++)=0x34; *(dd++)=0x12;}
					else {*(dd++)=r[c]*255.4; *(dd++)=g[c]*255.4; *(dd++)=b[c]*255.4;}
				}
			}
			#endif
		}
		else
		{
			#ifdef SBCLIB_GFX_D3D8
			// Don't do anything: there's no more to write!
			#elif defined(SBCLIB_GRAPHICS_OPENGL)
			#else
			if (cpp==1)
			{
				f->letter[n]=(unsigned char *)mallocex(3,"Letter off EOF");
				f->letter[n][0]=f->letter[n][1]=f->letter[n][2]=1;
			}
			else
			{
				f->letter[n]=(unsigned char *)mallocex(5,"Letter off EOF");
				f->letter[n][0]=f->letter[n][1]=1; f->letter[n][2]=f->letter[n][3]=0; f->letter[n][4]=128;
			}
			#endif
		}
	}
	#ifdef SBCLIB_GFX_D3D8
	f->letter->UnlockRect(0);
	#elif defined(SBCLIB_GRAPHICS_OPENGL)
	glGenTextures(1,&f->letter);
	glBindTexture(GL_TEXTURE_2D,f->letter);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_LOD_BIAS,-0.5);
	//glTexImage2D(GL_TEXTURE_2D,0,4,f->txs,f->tys,0,GL_RGBA,GL_UNSIGNED_BYTE,ta);
	sjbBuild2DMipmaps(4,f->txs,f->tys,ta); // About 2.5x faster than the glu version
	//gluBuild2DMipmaps(GL_TEXTURE_2D,4,f->txs,f->tys,GL_RGBA,GL_UNSIGNED_BYTE,ta);
	free(ta);
	correctmipmapgamma(f);
	#endif
	f->v=BF_variation_none; f->v.from=f; f->v.render_c=512;
	BF_fontlookups(f);
	BF_initialise();
	LSet_add(&BF_fontset,&f);
	if (!BF_curfont) bfsetfontpointer(f);
}

void loadfont(const char *filename,const char *handle)
{
	unsigned m; unsigned char *d=filedata(filename,&m);
	if (d)
	{
		loadfontdata(d,m,handle);
		free(d);
	}
	else // Hard failure, rather than crash later
	{
		char str[300]; sprintf(str,"loadfont: '%s' not found",filename);
		reportfatala(str);
	}
}

#define loadfontresource(R,H) loadfontdata(R,sizeof(R),H)
#define loadfontresourceRLE(R,H) \
{ \
	unsigned m=sizeof(R); unsigned char *tmp=(unsigned char *)malloc(m); \
	memcpy(tmp,R,m); \
	unrle(&tmp,&m); \
	loadfontdata(tmp,m,H); \
	free(tmp); \
} // Requires <shrink.c>

#ifndef SBCLIB_GFX_D3D8 // Not supported (yet)
#ifndef SBCLIB_GRAPHICS_OPENGL
void savefont(const char *filename,const char *handle)
{
	int n; FILE *out;
	BF_font *f=NULL;
	for (n=0;n<BF_fontset.m;n++) if (!strcmpi(handle,BF_array[n]->name)) f=BF_array[n];
	if (!f) return;
	out=fopen(filename,"wb");
	if (!out) return;
	if (cpp==1) for (n=0;n<512;n++)
		fwrite(f->letter[n],2+f->letter[n][0]*f->letter[n][1],1,out);
	else
	{
		int i; unsigned char *p;
		for (n=0;n<512;n++)
		{
			fwrite(f->letter[n],2,1,out); p=f->letter[n]+2;
			for (i=f->letter[n][0]*f->letter[n][1]-1;i>=0;i--)
			{
				fputc(40+p[0]/43+p[1]/43*6+p[2]/43*36,out);
				p+=3;
			}
		}
	}
	fclose(out);
}
#endif
#endif

void deletefont(const char *handle) {if (BF_isfont(handle)) BF_deletefont(bffont(handle));}

void deletefontvariants(const char *handle) {if (BF_isfont(handle)) BF_deletefontvariants(bffont(handle));}

char *bfstrend(char *s)
{
	while (*s) switch (*s)
	{
		case 0x02: s+=3; break;
		case 0x05: s+=6; break;
		default: s++; break;
	}
	return s;
}

/* Greek is chars 256-279 lowercase, in alphabetical order, 280-303 uppercase
304-383 reserved for mathematical and other symbols
Hebrew alphabet is 384-410
53 random CJK characters are in 411-463: U+4E02,4E06..4ED2 */

char *greekencode(char *s)
{ // In-place: make sure you've got room in your string
	const char *gname[24]={"alpha","beta","gamma","delta","epsilon","zeta","eta","theta","iota","kappa","lambda","mu",
		"nu","xi","omicron","pi","rho","sigma","tau","upsilon","phi","chi","psi","omega"};
	typedef struct {const char *s; int c;} Mapping;
	Mapping gnome[]={{"vardelta",304},{"varzeta",305},{"vartheta",306},{"varphi",307},{"varomega",308}};
	int n; char *t,*u;
	for (t=s;*t;t++)
	{
		for (n=0;n<24;n++) if (striseg(t,gname[n]))
		{
			unsigned short code=256+n+24*(*t<'a');
			int i=strlen(gname[n])-3;
			if (i>0) {for (u=t;u[i];u++) *u=u[i]; *u=u[i];}
			else if (i<0) for (u=bfstrend(t);u>t;u--) u[-i]=*u;
			t[0]=0x02; t[1]=code>>8; t[2]=code&0xFF;
			t+=2; n=24;
		}
		for (n=sizeof(gnome)/sizeof(*gnome)-1;n>=0;n--) if (striseg(t,gnome[n].s))
		{
			unsigned short code=gnome[n].c;
			int i=strlen(gnome[n].s)-3;
			if (i>0) {for (u=t;u[i];u++) *u=u[i]; *u=u[i];}
			else if (i<0) for (u=bfstrend(t);u>t;u--) u[-i]=*u;
			t[0]=0x02; t[1]=code>>8; t[2]=code&0xFF;
			t+=2; n=-1;
		}
	}
	return s;
}

char *hashencode(char *s)
{ // In-place again ("#col=123#" --> 05 00 0000007B)
	int n,p; char str[100],*t; const char *prefix[2]={"col","tab"};
	for (t=s;*t;t++)
	{
		for (p=0;p<2;p++)
		{
			sprintf(str,"#%s=",prefix[p]);
			if (striseg(t,str))
			{
				char *u=t,*v; n=0;
				for (t+=strlen(str);*t!='#' && *t;t++) n=n*10+*t-'0';
				u[0]=0x05; u[1]=p; memcpy(u+2,&n,sizeof(int)); t+=(*t=='#');
				for (v=u+2+sizeof(int);*t;v++,t++) *v=*t; *v=0;
				t=u+2+sizeof(int)-1;
				p=999;
			}
		}
	}
	return s;
}

#include <ctype.h>

void subsuperencode(char *s,const int level=-1,const short baseline=0,const short size=0x100)
{ // In-place... perhaps a little risky, so use a large buffer
	if (level==0) return;
	short subsize=//Max(size*0.7,0x80); // Exponential sequence 1, 0.7, 0.49... cut off at 0.5
		(short)(0.5+256.0*7*size/(256.0*7+3.0*size)); // Harmonic sequence 7/7, 7/10, 7/13...
	const float super=0.3,sub=-0.3,blpos=0.5;
	char *t;
	for (t=s;*t;t++) if (*t=='^' || *t=='_')
	{
		char *u=t+1,*v,*w; int brack=0;
		if (isalpha(*u)) while (isalpha(*u)) u++;
		else if (isdigit(*u)) while (isdigit(*u)) u++;
		else if (*u==0x02) while (*u==0x02) u+=3;
		else if (*u=='{')
		{
			int d=1; u++;
			while (d && *u)
			{
				if (*u=='{') d++; else if (*u=='}') d--;
				if (*u==0x02) u+=3; else u++;
			}
			u--; brack=1;
			int l1=u-(t+2),l2;
			char *b=(char *)malloc((l1+1)*2+100); memcpy(b,t+2,l1); b[l1]=0;
			if (*t=='^') subsuperencode(b,level-1,baseline+(short)(size*super-blpos*(size-subsize)),subsize);
			else subsuperencode(b,level-1,baseline+(short)(size*sub-blpos*(size-subsize)),subsize);
			char *be=bfstrend(b); l2=be-b;
			for (v=bfstrend(u);v>=u;v--) v[l2-l1]=*v;
			memcpy(t+2,b,l2); free(b);
			u+=l2-l1;
		}
		else u++;
		for (v=bfstrend(t),w=v+5-brack;v>t;v--,w--) *w=*v; u+=5-brack;
		if (*t=='^') *(short *)(t+4)=(short)(baseline+size*super-blpos*(size-subsize));
		else *(short *)(t+4)=(short)(baseline+size*sub-blpos*(size-subsize));
		t[0]=0x05; t[1]=0x02;
		*(unsigned short *)(t+2)=subsize;
		for (v=bfstrend(u),w=v+6-brack;v>=u;v--,w--) *w=*v;
		u[0]=0x05; u[1]=0x02;
		*(unsigned short *)(u+2)=size;
		*(short *)(u+4)=baseline;
		t=u+5;
	}
	else if (*t==0x02) t+=2;
}

char *eqnencode(char *s,const int level=-1)
{ // In-place... perhaps a little risky, so use a large buffer
	greekencode(s);
	subsuperencode(s,level,0,0x100);
	return s;
}

#ifdef SBCLIB_GRAPHICS_OPENGL
#include <gl/shaderprocs.c>
float bfwritewrap_scalex=1,bfwritewrap_scaley=1; int bfwrite_use_ortho=1;
#define bfwrites(ox,y,sc,s,c) (bfwritewrap_scalex=bfwritewrap_scaley=(sc),bfwritewrap((ox)/bfwritewrap_scalex,(y)/bfwritewrap_scaley,0,0,s,c),bfwritewrap_scalex=bfwritewrap_scaley=1)
#define bfwriteswrap(ox,y,xw,yw,sc,s,c) (bfwritewrap_scalex=bfwritewrap_scaley=(sc),bfwritewrap((ox)/bfwritewrap_scalex,(y)/bfwritewrap_scaley,(xw)/bfwritewrap_scalex,(yw)/bfwritewrap_scaley,s,c),bfwritewrap_scalex=bfwritewrap_scaley=1)
#define bfwriteh(ox,y,h,s,c) bfwrites(ox,y,(float)h/BF_height,s,c)
#define bfwritehwrap(ox,y,xw,yw,h,s,c) bfwriteswrap(ox,y,xw,yw,(float)h/BF_height,s,c)
#endif
void bfwrite_begin(void)
{ // Graphics initialisation for writing with a particular font
	#ifdef SBCLIB_GFX_D3D8
	typedef struct {float x,y,z; unsigned c; float u,v;} BF_Vertex;
	d3dev->SetVertexShader(D3DFVF_XYZ|D3DFVF_DIFFUSE|D3DFVF_TEX1);
	DWORD rs[5];
	d3dev->GetRenderState(D3DRS_ALPHABLENDENABLE,&rs[0]);
	d3dev->GetRenderState(D3DRS_LIGHTING,&rs[1]);
	d3dev->GetRenderState(D3DRS_ZWRITEENABLE,&rs[2]);
	d3dev->GetRenderState(D3DRS_SRCBLEND,&rs[3]);
	d3dev->GetRenderState(D3DRS_DESTBLEND,&rs[4]);
	d3dev->SetRenderState(D3DRS_ALPHABLENDENABLE,TRUE);
	d3dev->SetRenderState(D3DRS_LIGHTING,FALSE);
	d3dev->SetRenderState(D3DRS_ZWRITEENABLE,FALSE);
	switch (BF_paint)
	{
		case Solid:
			d3dev->SetRenderState(D3DRS_SRCBLEND,D3DBLEND_ONE);
			d3dev->SetRenderState(D3DRS_DESTBLEND,D3DBLEND_ZERO); break;
		case And:
			d3dev->SetRenderState(D3DRS_SRCBLEND,D3DBLEND_DSTCOLOR);
			d3dev->SetRenderState(D3DRS_DESTBLEND,D3DBLEND_INVSRCALPHA); break;
		case Or:
			d3dev->SetRenderState(D3DRS_SRCBLEND,D3DBLEND_ONE);
			d3dev->SetRenderState(D3DRS_DESTBLEND,D3DBLEND_INVSRCALPHA); break;
		case Alpha:
			d3dev->SetRenderState(D3DRS_SRCBLEND,D3DBLEND_SRCALPHA);
			d3dev->SetRenderState(D3DRS_DESTBLEND,D3DBLEND_INVSRCALPHA); break;
	}
	if (bfwrite_use_ortho)
	{
		D3DMATRIX oworld,oview,oproj;
		d3dev->GetTransform(D3DTS_WORLD,&oworld);
		d3dev->GetTransform(D3DTS_VIEW,&oview);
		d3dev->GetTransform(D3DTS_PROJECTION,&oproj);
		d3dev->SetTransform(D3DTS_WORLD,&D3DMATRIX_IDENTITY);
		d3dev->SetTransform(D3DTS_VIEW,&D3DMATRIX_IDENTITY);
		int j; D3DMATRIX proj; for (i=0;i<4;i++) for (j=0;j<4;j++) proj.m[i][j]=(i==j);
		proj._11=2.0/xres; proj._22=-2.0/yres;
		proj._41=-1; proj._42=+1;
		d3dev->SetTransform(D3DTS_PROJECTION,&proj);
	}
	IDirect3DTexture8 *otex;
	d3dev->GetTexture(0,&otex);
	d3dev->SetTexture(0,BF_curfont->letter);
	#elif defined(SBCLIB_GRAPHICS_OPENGL)
	glPushAttrib(GL_TRANSFORM_BIT|GL_ENABLE_BIT|GL_COLOR_BUFFER_BIT|GL_TEXTURE_BIT);
	if (bfwrite_use_ortho) ortho_on();
	glScalef(bfwritewrap_scalex,bfwritewrap_scaley,1);
	glEnable(GL_BLEND);
	glDisable(GL_LIGHTING);
	switch (BF_paint)
	{
		case Solid: glBlendFunc(GL_ONE,GL_ZERO); break;
		case And: glBlendFunc(GL_DST_COLOR,GL_ONE_MINUS_SRC_ALPHA); break;
		case Or: glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_COLOR); break;
		case Alpha: glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA); break;
	}
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D,BF_curfont->letter);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
	if (BF_shaderprog) glUseProgram(BF_shaderprog);
	glBegin(GL_QUADS);
	#endif
}

void bfwrite_end(void)
{
	#ifdef SBCLIB_GFX_D3D8
	d3dev->SetRenderState(D3DRS_ALPHABLENDENABLE,rs[0]);
	d3dev->SetRenderState(D3DRS_LIGHTING,rs[1]);
	d3dev->SetRenderState(D3DRS_ZWRITEENABLE,rs[2]);
	d3dev->SetRenderState(D3DRS_SRCBLEND,rs[3]);
	d3dev->SetRenderState(D3DRS_DESTBLEND,rs[4]);
	if (bfwrite_use_ortho)
	{
		d3dev->SetTransform(D3DTS_WORLD,&oworld);
		d3dev->SetTransform(D3DTS_VIEW,&oview);
		d3dev->SetTransform(D3DTS_PROJECTION,&oproj);
	}
	d3dev->SetTexture(0,otex);
	#elif defined(SBCLIB_GRAPHICS_OPENGL)
	glEnd();
	if (BF_shaderprog) glUseProgram(0);
	if (bfwrite_use_ortho) ortho_off();
	glPopAttrib();
	#endif
}

float bfwritewrap_inner(const float ox,float y,const float xw,const float yw,const char *s,unsigned c)
{ // Accepts 2-byte widechars with escapes: 02 ww ww and other commands 05 xx ii ii ii ii
	float bflength(const char *); // Defuled lower down
	if (!s) s="NULL?";
	if (!*s) return y; // Don't even do the setup for a zero-length string
	int n,i,
	#ifdef SBCLIB_GRAPHICS_OPENGL
		clipmargin=0; float x=ox,nx;
	#else
		x=ox,nx,clipmargin=xmin+BF_leftclipmargin;
	#endif
	BF_variation ov=BF_curfont->v,v=ov;
	int spacing=(int)(0.5+BF_height*BF_spacmul);
	int yoffset=0;
	#ifdef SBCLIB_GFX_D3D8
	BF_Vertex vert[4]; float xs,ys,txscal=1.0/BF_curfont->txs;
	vert[0].c=vert[1].c=vert[2].c=vert[3].c=c;
	#elif defined(SBCLIB_GRAPHICS_OPENGL)
	float txscal=1.0/BF_curfont->txs,tyscal=1.0/BF_curfont->tys,tx0,tx1,ty0,ty1,lxs,lh,xs,ys,bor;
	unsigned char calpha=0xFF;
	glColor4ub(c>>16,(c>>8)&0xFF,c&0xFF,calpha);
	#endif
	for (n=0;s[n];n++)
	{
		int sn=c2i(s[n]);
		if (sn==0x02) {sn=c2i(s[n+1])*256+c2i(s[n+2]); n+=2;}
		else if (sn==0x05)
		{
			int p=c2i(s[n+1]);
			if (p==0x00) // col
			{
				memcpy(&c,s+n+2,sizeof(unsigned));
				#ifdef SBCLIB_GRAPHICS_OPENGL
				glColor4ub(c>>16,(c>>8)&0xFF,c&0xFF,calpha);
				#endif
			}
			else if (p==0x01) // tab
				{int xnew; memcpy(&xnew,s+n+2,sizeof(int)); x=xnew;}
			else if (p==0x02) // super/sub script: change scale and baseline height
			{ // 05 02 0100 0000 resumes default
				float sc=*(unsigned short *)(s+n+2)/256.0;
				v.xscale=ov.xscale*sc; v.yscale=ov.yscale*sc;
				y+=yoffset;
				yoffset=(int)((*(short *)(s+n+4))*(float)ov.from->height*ov.yscale+128)/256;
				//if (KEY(VK_SHIFT) && ov.from->height>25) yoffset*=0.75;
				y-=yoffset;
				#ifdef SBCLIB_GFX_OPENGL
				glEnd();
				#endif
				#ifndef SBCLIB_GFX_D3D8
				bfsetfontvar(v);
				#endif
				#ifdef SBCLIB_GFX_OPENGL
				glBegin(GL_QUADS);
				#endif
			}
			n+=1+sizeof(int); goto next_char;
		}
		if (sn=='\n') {x=ox; y+=(int)(BF_curfont->height*v.yscale+0.5)+1;}
		else
		{
			#if defined(SBCLIB_GFX_D3D8) || defined(SBCLIB_GRAPHICS_OPENGL)
			lxs=(float)v.from->lxs[sn]-2.0*v.from->border;
			#else
			lxs=(float)v.from->letter[sn][0]-2.0*v.from->border;
			#endif
			lh=(float)v.from->height-2.0*v.from->border;
			#ifdef SBCLIB_GFX_D3D8
			if (BF_mono) {nx=x+BF_monospacing; x+=(BF_monospacing-lxs)/2;}
			else nx=x+(int)(0.5+lxs*v.xscale*(1.0+v.boldness))+spacing;
			#elif defined(SBCLIB_GRAPHICS_OPENGL)
			if (BF_mono) {nx=x+BF_monospacing; x+=(BF_monospacing-lxs)/2;}
			else
			{
				if (!BF_shaderprog) nx=x+(int)(0.5+lxs*v.xscale*(1.0+v.boldness)+BF_height*BF_spacmul);
				else nx=x+lxs*v.xscale*(1.0+v.boldness)+BF_height*BF_spacmul;
			}
			#else
			if (BF_mono) {nx=x+BF_monospacing; x+=(BF_monospacing-lxs)/2;}
			else nx=x+(int)(0.5+lxs*v.xscale*(1.0+v.boldness))+spacing; // Laptop LCC's optimise switch causes errors here
			#endif
			if (xw && sn==' ')
			{
				for (i=1;s[n+i] && !isspace(s[n+i]);i++);
				char *t=(char *)malloc(i); memcpy(t,s+n+1,i-1); t[i-1]=0;
				if (x+bflength(t)>xw) {nx=ox; y+=lh*v.yscale+1; if (yw && y>yw) break;}
				free(t);
			}
			if (sn!=' ')
			{
				#ifdef SBCLIB_GFX_D3D8
				xs=lxs*v.xscale;
				ys=lh*v.yscale;
				vert[0].x=x-0.5; vert[0].y=y-0.5; vert[0].z=0;
				vert[1].x=x+xs-0.5; vert[1].y=y-0.5; vert[1].z=0;
				vert[2].x=x-0.5; vert[2].y=y+ys-0.5; vert[2].z=0;
				vert[3].x=x+xs-0.5; vert[3].y=y+ys-0.5; vert[3].z=0;
				vert[0].u=(BF_curfont->lox[sn])*txscal; vert[0].v=0;
				vert[1].u=vert[0].u+BF_curfont->lxs[sn]*txscal; vert[1].v=0;
				vert[2].u=vert[0].u; vert[2].v=1;
				vert[3].u=vert[0].u+BF_curfont->lxs[sn]*txscal; vert[3].v=1;
				d3dev->DrawPrimitiveUP(D3DPT_TRIANGLESTRIP,2,vert,sizeof(BF_Vertex));
				#elif defined(SBCLIB_GRAPHICS_OPENGL)
				tx0=(BF_curfont->lox[sn])*txscal; // Do 0132 compared to above order
				tx1=tx0+BF_curfont->lxs[sn]*txscal;
				ty0=(BF_curfont->loy[sn])*tyscal;
				ty1=ty0+BF_curfont->height*tyscal;
				xs=lxs*v.xscale; /**/ // May need to use font variation parameters here, not in the overall glScale
				ys=lh*v.yscale;
				bor=v.from->border*v.from->height/lh;
				glTexCoord2f(tx0,ty0); glVertex2f(x-bor*v.xscale,y-bor*v.yscale);
				glTexCoord2f(tx1,ty0); glVertex2f(x+xs+bor*v.xscale,y-bor*v.yscale);
				glTexCoord2f(tx1,ty1); glVertex2f(x+xs+bor*v.xscale,y+ys+bor*v.yscale);
				glTexCoord2f(tx0,ty1); glVertex2f(x-bor*v.xscale,y+ys+bor*v.yscale);
				#else
				if (x>=clipmargin && y>=ymin && nx<=xres && y+BF_height<=yres) paintbitc(x,y,BF_letter[sn],c);
				else if (x>=xres || y>=yres || nx<=0 || y+BF_height<=0) ; // Offscreen!
				else if (cpp==1)
				{
					int px,py,xs=BF_letter[sn][0],ys=BF_letter[sn][1];
					unsigned char *p=BF_letter[sn]+2;
					for (py=0;py<ys;py++) for (px=0;px<xs;px++,p++)
						if (x+px>=clipmargin && x+px<xres && y+py>=xmin && y+py<yres && *p && *p!=255)
							paintpix(x+px,y+py,BF_and[c][*p]);
				}
				else
				{
					int px,py,xs=BF_letter[sn][0],ys=BF_letter[sn][1];
					unsigned char *p=BF_letter[sn]+2;
					int r1=c>>16,g1=(c>>8)&0xFF,b1=c&0xFF;
					for (py=0;py<ys;py++) for (px=0;px<xs;px++,p+=3)
						if (x+px>=clipmargin && x+px<xres && y+py>=xmin && y+py<yres && (p[0]!=0x56 || p[1]!=0x34 || p[2]!=0x12))
							paintpix(x+px,y+py,(((r1*p[2])>>8)<<16)|(((g1*p[1])>>8)<<8)|((b1*p[0])>>8));
				}
				#endif
			}
			if (xw && sn!=' ' && s[n+1]!=' ' && nx>xw) {x=ox; y+=BF_curfont->height+1; if (yw && y>yw) break;}
			else x=nx;
		}
		next_char:;
	}
	#ifndef SBCLIB_GFX_D3D8
	bfsetfontvar(ov);
	#endif
	return y;
}

float bfwritewrap(const float ox,const float y,const float xw,const float yw,const char *s,const unsigned c)
{
	bfwrite_begin();
	float ret=bfwritewrap_inner(ox,y,xw,yw,s,c);
	bfwrite_end();
	return ret;
}
// For many bfwrites the _inner version is 8x faster (the all-in-one perhaps 15% slower than direct)
#define bfwrite(ox,y,s,c) bfwritewrap(ox,y,0,0,s,c)
#define bfwrite_inner(ox,y,s,c) bfwritewrap_inner(ox,y,0,0,s,c)

float bflength_char(const int i)
{
	if (BF_mono) return BF_monospacing;
	else if (BF_curfont)
	{
		#if defined(SBCLIB_GFX_D3D8) || defined(SBCLIB_GRAPHICS_OPENGL)
		float lxs=(float)BF_curfont->v.from->lxs[i]-2.0*BF_curfont->v.from->border;
		#else
		float lxs=(float)BF_curfont->v.from->letter[i][0]-2.0*BF_curfont->v.from->border;
		#endif
		if (!BF_shaderprog)
		{
			int spacing=(int)(0.5+(float)BF_height*BF_spacmul);
			return (int)(0.5+lxs*BF_curfont->v.xscale*(1.0+BF_curfont->v.boldness))+spacing;
		}
		else return lxs*BF_curfont->v.xscale*(1.0+BF_curfont->v.boldness)+BF_height*BF_spacmul;
	}
	else return 5;
}

float bflength__old(const char *s)
{
	int i; float x;
	for (x=0;*s && *s!='\n';s++)
	{
		i=c2i(*s);
		if (i<8) s+=i;
		else x+=bflength_char(i);
	}
	return x;
}

float bflength(const char *s)
{ // Accepts 2-byte widechars with escapes: 02 ww ww and other commands 05 xx ii ii ii ii
	if (!s) s="NULL?";
	if (!*s) return 0; // Don't even do the setup for a zero-length string
	int n,i; float x=0,nx;
	BF_variation v=BF_curfont->v; float vxscale=v.xscale,lxs;
	for (n=0;s[n];n++)
	{
		int sn=c2i(s[n]);
		if (sn==0x02) {sn=c2i(s[n+1])*256+c2i(s[n+2]); n+=2;}
		else if (sn==0x05)
		{
			int p=c2i(s[n+1]);
			if (p==0x00) ; // col
			else if (p==0x01) // tab
				{int xnew; memcpy(&xnew,s+n+2,sizeof(int)); x=xnew;}
			else if (p==0x02) // super/sub script: change scale and baseline height
			{ // 05 02 0100 0000 resumes default
				float sc=*(unsigned short *)(s+n+2)/256.0;
				vxscale=v.xscale*sc;
			}
			n+=1+sizeof(int); goto next_char;
		}
		if (sn=='\n') ;
		else
		{
			lxs=(float)v.from->lxs[sn]-2.0*v.from->border;
			if (BF_mono) {nx=x+BF_monospacing; x+=(BF_monospacing-lxs)/2;}
			else
			{
				if (!BF_shaderprog) nx=x+(int)(0.5+lxs*vxscale*(1.0+v.boldness)+BF_height*BF_spacmul);
				else nx=x+lxs*vxscale*(1.0+v.boldness)+BF_height*BF_spacmul;
			}
			x=nx;
		}
		next_char:;
	}
	return x;
}

// Check these two are returning the correct thing (no trailing spacer)
float bflengtht(char *s,const int n)
{
	if (n>=0 && n<strlen(s))
	{
		char t=s[n]; s[n]=0;
		float ret=bflength(s);
		s[n]=t; return ret;
	}
	else return bflength(s);
}

int bfclosestgap(const char *s,const int z)
{ // Should return character gap number that is closest to offset x
	int n,i,x,y;
	for (n=x=0;s[n] && s[n]!='\n';n++)
	{
		i=c2i(s[n]);
		if (i<8) n+=i;
		else
		{
			y=x+bflength_char(i);
			if (y>=z) {if (z-x<y-z) return n; else return n+1;}
			x=y;
		}
	}
	return n;
}

void bfwritebg(const int x,const int y,char *s,const unsigned cf,const unsigned cb=0)
{
	rect(x,y,bflength(s),BF_height,cb);
	bfwrite(x,y,s,cf);
}

#if !defined(SBCLIB_GRAPHICS_OPENGL) && !defined(SBCLIB_GRAPHICS_D3D8)

void editfont(void)
{ /**/
	int n,i,j,x,y;
	int cur=0,col=0;
	char leave=0;
	while (!leave)
	{
		int xs=BF_letter[cur][0],ys=BF_letter[cur][1];
		int xspacing=Mini(BF_width+1,(xres/2-160)/16),yspacing=Mini(BF_height+1,yres/32);
		float zoom=Min((float)xres/2/xs,(float)yres/ys);
		BF_curfont->width=BF_width=BF_curfont->height=BF_height=0;
		for (n=0;n<512;n++)
		{
			BF_curfont->width=BF_width=Max(BF_width,BF_letter[n][0]);
			BF_curfont->height=BF_height=Max(BF_height,BF_letter[n][1]);
		}
		for (n=0;n<512;n++) if (n!=cur || GetTickCount()/250%2==0)
			wrapbit(xres-(16-n%16)*xspacing,yres-(32-n/16)*yspacing,BF_letter[n]);
		for (n=0;n<256;n++) rect(xres/2+1+n%16*10,n/16*10+1,9,9,n);
		box(xres/2+col%16*10,col/16*10,10,10,7+GetTickCount()/100%2*8);
		for (x=0;x<xs;x++) for (y=0;y<ys;y++)
		{
			unsigned char c=BF_letter[cur][2+x+y*xs];
			if (c!=255) rect(x*zoom,y*zoom,zoom+1,zoom+1,c);
			else box(x*zoom+zoom/4,y*zoom+zoom/4,zoom/2,zoom/2,1);
		}
		if (mb&1)
		{
			if (mx<zoom*xs && my<zoom*ys)
				BF_letter[cur][2+(int)floor(mx/zoom)+(int)floor(my/zoom)*xs]=col;
			else if (mx>=xres/2 && mx<xres/2+160 && my<160)
				col=(mx-xres/2)/10+my/10*16;
			else if (mx>=xres-xspacing*16 && my>=yres-yspacing*32)
				cur=(mx-xres+xspacing*16)/xspacing+(my-yres+yspacing*32)/yspacing*16;
		}
		if (mb&2)
		{
			if (mx<zoom*xs && my<zoom*ys)
				col=BF_letter[cur][2+(int)floor(mx/zoom)+(int)floor(my/zoom)*xs];
		}
		if (KEY(VK_UP) && ys>1)
		{
			unsigned char *bit=(unsigned char *)mallocex(xs*ys,"bit#1, editfont(), B.F.");
			memcpy(bit,BF_letter[cur]+2,xs*ys);
			free(BF_letter[cur]);
			ys--;
			#ifdef SBCLIB_DEBUG
			char str[100]; sprintf(str,"BF_letter code %d, alloc'd by editfont[UP]",cur);
			#endif
			BF_letter[cur]=(unsigned char *)mallocex(2+xs*ys,str);
			BF_letter[cur][0]=xs; BF_letter[cur][1]=ys;
			for (x=0;x<xs;x++) for (y=0;y<ys;y++)
				BF_letter[cur][2+x+y*xs]=bit[x+y*xs];
			free(bit);
			KEYUP(VK_UP);
		}
		if (KEY(VK_DOWN) && ys<100)
		{
			unsigned char *bit=(unsigned char *)mallocex(xs*ys,"bit#2, editfont(), B.F.");
			memcpy(bit,BF_letter[cur]+2,xs*ys);
			free(BF_letter[cur]);
			ys++;
			#ifdef SBCLIB_DEBUG
			char str[100]; sprintf(str,"BF_letter code %d, alloc'd by editfont[DOWN]",cur);
			#endif
			BF_letter[cur]=(unsigned char *)mallocex(2+xs*ys,str);
			BF_letter[cur][0]=xs; BF_letter[cur][1]=ys;
			for (x=0;x<xs;x++)
			{
				for (y=0;y<ys-1;y++)
					BF_letter[cur][2+x+y*xs]=bit[x+y*xs];
				BF_letter[cur][2+x+(ys-1)*xs]=255;
			}
			free(bit);
			KEYUP(VK_DOWN);
		}
		if (KEY(VK_LEFT) && xs>1)
		{
			unsigned char *bit=(unsigned char *)mallocex(xs*ys,"bit#3, editfont(), B.F.");
			memcpy(bit,BF_letter[cur]+2,xs*ys);
			free(BF_letter[cur]);
			xs--;
			#ifdef SBCLIB_DEBUG
			char str[100]; sprintf(str,"BF_letter code %d, alloc'd by editfont[LEFT]",cur);
			#endif
			BF_letter[cur]=(unsigned char *)mallocex(2+xs*ys,str);
			BF_letter[cur][0]=xs; BF_letter[cur][1]=ys;
			for (x=0;x<xs;x++) for (y=0;y<ys;y++)
				BF_letter[cur][2+x+y*xs]=bit[x+y*(xs+1)];
			free(bit);
			KEYUP(VK_LEFT);
		}
		if (KEY(VK_RIGHT) && xs<100)
		{
			unsigned char *bit=(unsigned char *)mallocex(xs*ys,"bit#4, editfont(), B.F.");
			memcpy(bit,BF_letter[cur]+2,xs*ys);
			free(BF_letter[cur]);
			xs++;
			#ifdef SBCLIB_DEBUG
			char str[100]; sprintf(str,"BF_letter code %d, alloc'd by editfont[RIGHT]",cur);
			#endif
			BF_letter[cur]=(unsigned char *)mallocex(2+xs*ys,str);
			BF_letter[cur][0]=xs; BF_letter[cur][1]=ys;
			for (x=0;x<xs-1;x++) for (y=0;y<ys;y++)
				BF_letter[cur][2+x+y*xs]=bit[x+y*(xs-1)];
			for (y=0;y<ys;y++) BF_letter[cur][2+(xs-1)+y*xs]=255;
			free(bit);
			KEYUP(VK_RIGHT);
		}
		if (KEY(VK_D) && xs<=50 && ys<=50)
		{
			unsigned char *bit=(unsigned char *)mallocex(xs*ys,"bit#5, editfont(), B.F.");
			memcpy(bit,BF_letter[cur]+2,xs*ys);
			free(BF_letter[cur]);
			xs*=2; ys*=2;
			#ifdef SBCLIB_DEBUG
			char str[100]; sprintf(str,"BF_letter code %d, alloc'd by editfont[D]",cur);
			#endif
			BF_letter[cur]=(unsigned char *)mallocex(2+xs*ys,str);
			BF_letter[cur][0]=xs; BF_letter[cur][1]=ys;
			for (x=0;x<xs/2;x++) for (y=0;y<ys/2;y++)
			{
				unsigned char c=bit[x+y*xs/2];
				BF_letter[cur][2+x*2+y*2*xs]=c;
				BF_letter[cur][2+(x*2+1)+y*2*xs]=c;
				BF_letter[cur][2+x*2+(y*2+1)*xs]=c;
				BF_letter[cur][2+(x*2+1)+(y*2+1)*xs]=c;
			}
			free(bit);
			KEYUP(VK_D);
		}
		#ifndef SBCLIB_GFX_D3D8
		if (KEY(VK_T)) BFF_trimfont(BF_curfont,0);
		#endif
		line(mx,my,mx+2,my,15); line(mx,my,mx,my+2,15); line(mx,my,mx+4,my+4,15);
		ccl(0);
		leave|=dealwithmessages()|KEY(VK_Q);
		BF_monospacing=BF_letter['C'][0]+1;
	}
}
#endif

#define GEBUGSTR(STR) {bfwrite(rnd(xres-100),rnd(yres-20),STR,(cpp==1?15:0xFFFFFF)); update();}
#define GEBUG(N) {char tmpstr[100]; sprintf(tmpstr,"%d",N); GEBUGSTR(tmpstr);}

#ifdef SBCLIB_GRAPHICS_OPENGL
void bfwrite3d_begin(void)
{
	glPushAttrib(GL_COLOR_BUFFER_BIT|GL_ENABLE_BIT|GL_TEXTURE_BIT);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D,BF_curfont->letter);
	glEnable(GL_BLEND);
	switch (BF_paint)
	{
		case Solid: glBlendFunc(GL_ONE,GL_ZERO); break;
		case And: glBlendFunc(GL_ZERO,GL_ONE_MINUS_SRC_COLOR); break;
		case Or: glBlendFunc(GL_SRC_ALPHA,GL_ONE); break;
		case Alpha: glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA); break;
	}
	glEnable(GL_ALPHA_TEST);
	glAlphaFunc(GL_GREATER,0); // Alpha testing works!!  Alpha=0 pixels now don't update the depth buffer
	if (BF_shaderprog) glUseProgram(BF_shaderprog);
	glBegin(GL_QUADS);
}

void bfwrite3d_end(void)
{
	glEnd();
	if (BF_shaderprog) glUseProgram(0);
	glPopAttrib();
}

//void textcoords(double x,double y,void *pdata,double *px,double *py,double *pz,unsigned *pc) // x,y, in units of BF_height
typedef void (*bfwritewarp_fn)(double,double,void *,double *,double *,double *,unsigned *);
void bfwritewarp(const char *ss,const bfwritewarp_fn f,void *pdata=NULL)
{
	double x=0,y=0,px,py,pz; unsigned c; unsigned char *s=(unsigned char *)ss;
	double txscal=1.0/BF_curfont->txs,tyscal=1.0/BF_curfont->tys, // Texture scaling
		tx0,tx1,ty0,ty1,xs, // Texture coords and letter width
		x0,x1,y0,y1; // Row-scaled corner coords
	bfwrite3d_begin();
	for (;*s;s++)
	{
		if (*s=='\n') {x=0; y+=1;}
		else if (*s=='\t') x++;
		else
		{
			xs=BF_curfont->lxs[*s];
			tx0=(BF_curfont->lox[*s]-0.5)*txscal; // Do 0132 compared to above order
			tx1=tx0+(xs+1)*txscal;
			ty0=(BF_curfont->loy[*s]-0.5)*tyscal;
			ty1=ty0+(BF_height+1)*tyscal;
			x0=x-0.5/BF_height;
			x1=x+(xs+0.5)/BF_height;
			y0=y-0.5/BF_height;
			y1=y+(BF_height+0.5)/BF_height;
			glTexCoord2f(tx0,ty0); f(x0,y0,pdata,&px,&py,&pz,&c);
			glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF); glVertex3f(px,py,pz);
			glTexCoord2f(tx1,ty0); f(x1,y0,pdata,&px,&py,&pz,&c);
			glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF); glVertex3f(px,py,pz);
			glTexCoord2f(tx1,ty1); f(x1,y1,pdata,&px,&py,&pz,&c);
			glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF); glVertex3f(px,py,pz);
			glTexCoord2f(tx0,ty1); f(x0,y1,pdata,&px,&py,&pz,&c);
			glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF); glVertex3f(px,py,pz);
			x+=(xs+1)/BF_height;//*(*s==' '?1:0.85);
		}
	}
	bfwrite3d_end();
}

void bfwrite3d_inner(double x,double y,double z,const char *ss,
	double dx,double dy,double dz, // Down (one line)
	double ax,double ay,double az) // Along
{ // Needs bfwrite3d_begin and _end around it
	double px=x,py=y,pz=z; unsigned char *s=(unsigned char *)ss;
	double txscal=1.0/BF_curfont->txs,tyscal=1.0/BF_curfont->tys, // Texture scaling
		tx0,tx1,ty0,ty1,xs; // Texture coords and letter width
	for (;*s;s++)
	{
		if (*s=='\n') {x+=dx; y+=dy; z+=dz; px=x; py=y; pz=z;}
		else if (*s=='\t') {px+=ax; py+=ay; pz+=az;}
		else
		{
			xs=BF_curfont->lxs[*s];
			tx0=(BF_curfont->lox[*s])*txscal; // Do 0132 compared to above order
			tx1=tx0+xs*txscal;
			ty0=(BF_curfont->loy[*s])*tyscal;
			ty1=ty0+BF_height*tyscal;
			xs/=BF_height;
			glTexCoord2f(tx0,ty0); glVertex3f(px,py,pz);
			glTexCoord2f(tx1,ty0); glVertex3f(px+ax*xs,py+ay*xs,pz+az*xs);
			glTexCoord2f(tx1,ty1); glVertex3f(px+ax*xs+dx,py+ay*xs+dy,pz+az*xs+dz);
			glTexCoord2f(tx0,ty1); glVertex3f(px+dx,py+dy,pz+dz);
			xs+=1.0/BF_height;
			px+=ax*xs; py+=ay*xs; pz+=az*xs;
		}
	}
}

double bflengthw(const char *ss)
{ // Returns fractional length in units of BF_height, compatible with bfwritewarp
	double x=0,xs; unsigned char *s=(unsigned char *)ss;
	for (;*s;s++)
	{
		if (*s=='\n') {}
		else if (*s=='\t') x++;
		else
		{
			xs=BF_curfont->lxs[*s];
			x+=xs/BF_height;//*(*s==' '?1:0.85);
		}
	}
	return x;
}

typedef struct {double x,y,f; unsigned c;} bfwritesc_data;
void textcoords_bfwritesc(double x,double y,bfwritesc_data *d,double *px,double *py,double *pz,unsigned *pc) // x,y, in units of BF_height
{
	*px=d->x+x*d->f;
	*py=d->y+y*d->f;
	*pz=0; *pc=d->c;
}

void bfwritesc(const double x,const double y,const double h,const char *s,const unsigned c)
{
	bfwritesc_data d; d.x=x; d.y=y;
	d.f=h;
	d.c=c;
	bfwritewarp(s,(bfwritewarp_fn)textcoords_bfwritesc,&d);
}

void bfwrite3dlabel(const double x,const double y,const double z,const char *s,const unsigned c,const double xs=1,const double ys=1)
{
	double mm[16],pm[16];
	glGetDoublev(GL_MODELVIEW_MATRIX,mm);
	glGetDoublev(GL_PROJECTION_MATRIX,pm);
	double
		x1=mm[0]*x+mm[4]*y+mm[8]*z+mm[12],
		y1=mm[1]*x+mm[5]*y+mm[9]*z+mm[13],
		z1=mm[2]*x+mm[6]*y+mm[10]*z+mm[14],
		d2=pm[11]*z1+pm[15],
		x2=pm[0]*x1/d2,
		y2=pm[5]*y1/d2;
	if (x2<-1 || x2>1 || y2<-1 || y2>1) return;
	int v[4]; glGetIntegerv(GL_VIEWPORT,v); int X=v[2],Y=v[3];
	double dx2=2.0/X,dy2=2.0/Y;
/* orig
	x2=floor((x2+1)/dx2)*dx2-1;
	y2=floor((y2+1)/dy2)*dy2-1;
*/
	x2=(floor(x2/dx2+10000)-10000)*dx2;
	y2=(floor(y2/dy2+10000)-10000)*dy2;
	x1=x2/pm[0]*d2; y1=y2/pm[5]*d2;
	x1-=mm[12]; y1-=mm[13]; z1-=mm[14];
	double
		nx=mm[0]*x1+mm[1]*y1+mm[2]*z1,
		ny=mm[4]*x1+mm[5]*y1+mm[6]*z1,
		nz=mm[8]*x1+mm[9]*y1+mm[10]*z1;
	double edy=-ys*BF_height*2.0/Y;
	edy*=d2/pm[5];
	double 
		dx=mm[1]*edy,
		dy=mm[5]*edy,
		dz=mm[9]*edy;
	double eax=xs*BF_height*2.0/X;
	eax*=d2/pm[0];
	double 
		ax=mm[0]*eax,
		ay=mm[4]*eax,
		az=mm[8]*eax;		
	glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
	bfwrite3d_begin();	
	bfwrite3d_inner(nx,ny,nz,s,dx,dy,dz,ax,ay,az);
	bfwrite3d_end();
}
#endif

#define SBCLIB_BITMAPFONTS
#endif

